# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from megatron.energon.dataset_config import load_config
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import BaseCoreDataset
from megatron.energon.metadataset.dataset_loader import DatasetLoader
from megatron.energon.metadataset.loader_interface import DatasetLoaderInterface
from megatron.energon.worker import WorkerConfig


@dataclass
class DatasetReference:
    path: Union[str, EPath]
    split_part: Optional[str] = None
    subflavor: Optional[str] = None
    subflavors: Optional[Dict[str, Any]] = None
    shuffle_over_epochs_multiplier: int = 1
    dataset_config: str = "dataset.yaml"
    split_config: str = "split.yaml"

    weight: float = 1.0

    _dataset: Optional[Union["Metadataset", DatasetLoader]] = None

    def prepare(self, parent_path: EPath):
        self.path = parent_path.absolute() / self.path
        if self.path.is_file():
            assert self.dataset_config == "dataset.yaml", "Must not set dataset_config"
            assert self.split_config == "split.yaml", "Must not set split_config"
            self._dataset = load_config(
                self.path,
                default_type=Metadataset,
                strict=True,
                default_kwargs=dict(parent_path=self.path.parent),
            )
        elif self.path.is_dir():
            self._dataset = DatasetLoader(
                path=self.path,
                split_part=self.split_part,
                subflavor=self.subflavor,
                subflavors=self.subflavors,
                dataset_config=self.dataset_config,
                split_config=self.split_config,
            )
        else:
            raise FileNotFoundError(self.path)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: Optional[WorkerConfig] = None,
        subflavor: Optional[str] = None,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: int = 1,
        **kwargs,
    ) -> List[Tuple[BaseCoreDataset, float]]:
        if self.subflavors is not None:
            subflavors = {**self.subflavors, **(subflavors or {})}
        return self._dataset.get_datasets(
            training=training,
            split_part=self.split_part or split_part,
            worker_config=worker_config,
            subflavor=subflavor or self.subflavor,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier
            * self.shuffle_over_epochs_multiplier,
            **kwargs,
        )


@dataclass
class MetadatasetMixer:
    """Internal mixer for the dataset."""

    datasets: List[DatasetReference]

    def prepare(self, parent_path: EPath):
        parent_path = parent_path.absolute()
        for dataset in self.datasets:
            dataset.prepare(parent_path)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: Optional[WorkerConfig] = None,
        subflavor: Optional[str] = None,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: int = 1,
        **kwargs,
    ) -> List[Tuple[BaseCoreDataset, float]]:
        sum_weight = sum(dataset.weight for dataset in self.datasets)
        return [
            (loaded_dataset, weight * dataset.weight / sum_weight)
            for dataset in self.datasets
            for loaded_dataset, weight in dataset.get_datasets(
                training=training,
                split_part=split_part,
                worker_config=worker_config,
                subflavor=subflavor,
                subflavors=subflavors,
                shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
                **kwargs,
            )
        ]


class Metadataset(DatasetLoaderInterface):
    """Main entry for metadataset."""

    _splits: Dict[str, MetadatasetMixer]

    def __init__(
        self,
        parent_path: Union[EPath, str],
        splits: Dict[str, MetadatasetMixer],
    ):
        """Create the metadataset"""
        parent_path = EPath(parent_path).absolute()
        self._splits = splits
        # Fix paths
        for split in splits.values():
            split.prepare(parent_path)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: Optional[WorkerConfig] = None,
        subflavor: Optional[str] = None,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: int = 1,
        **kwargs,
    ) -> List[Tuple[BaseCoreDataset, float]]:
        return self._splits[split_part].get_datasets(
            training=training,
            split_part=split_part,
            worker_config=worker_config,
            subflavor=subflavor,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
            **kwargs,
        )



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, Generator, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

import torch

from megatron.energon.flavors.base_dataset import MergedState, SavableDataset, State
from megatron.energon.rng import WorkerRng, WorkerRngMergedState, WorkerRngState
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


@dataclass
class BlendDatasetState(State):
    #: States of the sub datasets
    datasets: List[State]
    #: State of the worker rng
    rng: WorkerRngState


@dataclass
class BlendDatasetMergedState(MergedState):
    #: States of the sub datasets
    datasets: List[MergedState]
    #: State of the worker rng
    rng: WorkerRngMergedState


class BlendDataset(BaseWrapperDataset[T_sample], Generic[T_sample]):
    """
    This dataset wrapper blends multiple iterable datasets together give a weighting.
    The datasets may be infinite. This dataset is always infinite.
    """

    dataset_weights: Tuple[Tuple[SavableDataset[T_sample], float], ...]

    _worker_rng: WorkerRng

    def __init__(
        self,
        *dataset_weights: Tuple[SavableDataset[T_sample], float],
        worker_config: Optional[WorkerConfig] = None,
    ):
        """Construct a BlendDataset.

        Args:
            dataset_weights: Each argument should be a tuple of (dataset, weight) with a weight
                between 0 and 1. The output samples are sampled from the input datasets with the
                given probabilities.
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
        """
        super().__init__()
        self.worker_config = worker_config or global_worker_config
        self.dataset_weights = dataset_weights
        self._worker_rng = WorkerRng(self.worker_config)

    def __len__(self) -> int:
        # Gives an approximation of the number of samples. This is very incorrect (as the length
        # is weighted by the dataset weights).
        total = sum(weight for _, weight in self.dataset_weights)
        return int(
            sum(len(dataset) * weight / total for dataset, weight in self.dataset_weights)
        ) * len(self.dataset_weights)

    def _repeat(self, dataset: SavableDataset[T_sample]) -> Generator[T_sample, None, None]:
        while True:
            yield from dataset

    def __iter__(self) -> Iterator[T_sample]:
        datasets, weights = zip(*self.dataset_weights)
        dataset_iters = [self._repeat(dataset) for dataset in datasets]
        weights = torch.tensor(weights, dtype=torch.float32)
        probs = weights / weights.sum()

        while True:
            ds_idx = self._worker_rng.choice_idx(probs=probs)
            yield self._add_sample_restore_key(next(dataset_iters[ds_idx]), ds_idx)

    def save_state(self) -> BlendDatasetState:
        return BlendDatasetState(
            datasets=[d.save_state() for d, _weight in self.dataset_weights],
            rng=self._worker_rng.save_state(),
        )

    def merge_states(self, states: List[BlendDatasetState]) -> BlendDatasetMergedState:
        assert all(s is None or isinstance(s, BlendDatasetState) for s in states)
        assert all(s is None or len(s.datasets) == len(self.dataset_weights) for s in states)
        return BlendDatasetMergedState(
            datasets=[
                d.merge_states([None if s is None else s.datasets[ds_idx] for s in states])
                for ds_idx, (d, _) in enumerate(self.dataset_weights)
            ],
            rng=self._worker_rng.merge_states([None if s is None else s.rng for s in states]),
        )

    def restore_state(self, state: Optional[BlendDatasetMergedState]) -> None:
        if state is None:
            for dataset, _weight in self.dataset_weights:
                dataset.restore_state(None)
            self._worker_rng.restore_state(None)
        else:
            assert isinstance(state, BlendDatasetMergedState)
            assert len(state.datasets) == len(self.dataset_weights)
            for (dataset, _weight), dstate in zip(self.dataset_weights, state.datasets):
                dataset.restore_state(dstate)
            self._worker_rng.restore_state(state.rng)

    def verify_worker_config(self, worker_config: WorkerConfig) -> None:
        super().verify_worker_config(worker_config)
        for dataset, _weight in self.dataset_weights:
            dataset.verify_worker_config(worker_config)

    def can_restore_sample(self) -> bool:
        return all(dataset.can_restore_sample() for dataset, _weight in self.dataset_weights)

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_sample:
        ds_idx = index[0]
        assert isinstance(ds_idx, int)
        return self._add_sample_restore_key(
            self.dataset_weights[ds_idx][0].restore_sample(index[1:]), ds_idx
        )

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset_weights": [
                (dataset.config(), weight) for dataset, weight in self.dataset_weights
            ],
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"BlendDataset(dataset_weights={self.dataset_weights})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import List, Optional

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class SimilarityInterleavedSample(Sample):
    """Sample type for interleaved media such as text with images, but without image-text alignment.
    That alignment has to be assigned from the similarity matrix."""

    #: The images of the sequence
    images: List[torch.Tensor]
    #: The texts of the sequence
    texts: List[str]

    #: Similarity matrix between image and text entries in the sequence
    similarity_matrix: Optional[torch.Tensor] = None


class SimilarityInterleavedWebdataset(DefaultDecoderWebdataset[SimilarityInterleavedSample]):
    __sample_type__ = SimilarityInterleavedSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from megatron.energon.flavors.base_dataset import BaseCoreDataset
from megatron.energon.worker import WorkerConfig


class DatasetLoaderInterface(ABC):
    """General interface for a dataset loader."""

    @abstractmethod
    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: Optional[WorkerConfig] = None,
        subflavor: Optional[str] = None,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: int = 1,
        **kwargs,
    ) -> List[Tuple[BaseCoreDataset, float]]:
        """
        Calls :func:`megatron.energon.dataset_config.get_dataset_from_config` (loads the raw dataset)
        for all innermost datasets and resolves their relative weights to absolute weights.

        Args:
            training: If true, apply training randomization.
            split_part: Default split part to use.
            worker_config: Worker configuration to use. Defaults to `global_worker_config`.
            subflavor: Set the default subflavor for all datasets.
            subflavors: Set the default subflavors for all datasets.
            shuffle_over_epochs_multiplier: Multiply the inner datasets
                `shuffle_over_epochs(_multiplier)` by this factor. E.g. if the inner dataset
                has `shuffle_over_epochs_multiplier=2` and this function has
                `shuffle_over_epochs_multiplier=3`, the inner dataset will be shuffled
                over 6 epochs. Shuffling over `n` epochs guarantees that each sample is seen
                exactly `n` times in `n` epochs of the inner dataset. Use -1 for shuffling over
                an infinite number of epochs (effectively, this will draw shard slices with
                replacement).
            **kwargs: Additional arguments to the dataset constructor.

        Returns:
            The instantiated core datasets and their weights.
        """
        ...



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import List, Optional

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class MultiChoiceVQASample(Sample):
    """Sample type for visual question answering."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor
    #: The context/question for the image
    context: str

    #: The candidate answers.
    choices: Optional[List[str]] = None
    #: The index of the correct answer.
    correct_choice_idx: int = 0


class MultiChoiceVQAWebdataset(DefaultDecoderWebdataset[MultiChoiceVQASample]):
    __sample_type__ = MultiChoiceVQASample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class TextSample(Sample):
    """Sample type for simple text."""

    #: The text of the sample
    text: str


class TextWebdataset(DefaultDecoderWebdataset[TextSample]):
    __sample_type__ = TextSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Callable, Optional

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.crude import CrudeSample


@dataclass
class Cooker:
    """A cooker transforms a crude sample (simple dict) into a
    specific sample type inheriting from `Sample`.

    The `cook` method performs the transformation,
    the other fields are used to select the samples which this cooker
    can transform. If no filters are provided, the cooker will
    transform any sample.
    """

    cook: Callable[[dict], Sample]

    # If multiple of the following conditions are provided
    # then the sample must satisfy all of them.
    is_subflavor: Optional[str] = None
    has_subflavors: Optional[dict] = None
    condition: Optional[Callable[[dict], bool]] = None

    def is_match(self, crude_sample: CrudeSample) -> bool:
        if self.is_subflavor is not None:
            if crude_sample["__subflavor__"] != self.is_subflavor:
                return False

        if self.has_subflavors is not None:
            # Checks if the dict entries provided as a filter all match
            # the ones in the sample. The sample may have additional entries.
            for k, v in self.has_subflavors.items():
                if (
                    k not in crude_sample["__subflavors__"]
                    or crude_sample["__subflavors__"][k] != v
                ):
                    return False

        if self.condition is not None:
            if not self.condition(crude_sample):
                return False

        return True


def basic_sample_keys(crude_sample: dict) -> dict:
    """A convenience helper to extract the basic keys from a crude sample,
    which you will always need to forward to the cooked sample."""

    return {k: v for k, v in crude_sample.items() if k in Sample.__dataclass_fields__.keys()}



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import List, Optional

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class VQAOCRSample(Sample):
    """Sample type for visual question answering."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor

    #: The context/question for the image (VQA)
    context: str
    #: The text contained in the image (OCR)
    text: str

    #: The possible answers. Not set for testing. (VQA)
    answers: Optional[List[str]] = None
    #: The weights of the possible answers. Optionally available. (VQA)
    answer_weights: Optional[torch.Tensor] = None
    #: The bounding boxes of the words in the image (N, 4|5<x, y, w, h[, confidence]>) (OCR)
    words_boxes: Optional[torch.Tensor] = None
    #: The text contained in each word (N,) (OCR)
    words_text: Optional[List[str]] = None


class VQAOCRWebdataset(DefaultDecoderWebdataset[VQAOCRSample]):
    __sample_type__ = VQAOCRSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Generic, List, Optional, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import MergedState, SavableDataset, State
from megatron.energon.worker import WorkerConfig

T_sample = TypeVar("T_sample")
T_sample_out = TypeVar("T_sample_out")
T_sample_in = TypeVar("T_sample_in")


@dataclass
class BaseSingleWrapperState(State):
    """Base class for dataset states."""

    #: The class name of the dataset saving this state for assertion
    dataset_type: str

    #: State of the inner dataset
    dataset_state: State


@dataclass
class BaseSingleWrapperMergedState(MergedState):
    """Base class for dataset states."""

    #: The class name of the dataset saving/merging this state for assertion
    dataset_type: str

    #: State of the inner dataset
    dataset_state: MergedState


class BaseWrapperDataset(SavableDataset[T_sample], Generic[T_sample]):
    """Base class for dataset wrappers. All dataset wrappers should derive from this. A dataset
    wrapper takes one dataset and modifies its samples to make a new dataset. This can be for
    shuffling samples or applying custom functions to the data. Some wrappers only modify the
    length of the dataset or how it's repeated."""


class BaseSingleWrapperDataset(
    BaseWrapperDataset[T_sample_out], Generic[T_sample_in, T_sample_out]
):
    """Base class for dataset wrappers that wrap a single dataset. Provides default implementations
    for saving and restoring the dataset state."""

    dataset: SavableDataset[T_sample_in]

    def __init__(self, dataset: SavableDataset[T_sample_in]):
        super().__init__()
        self.dataset = dataset

    def save_state(self) -> BaseSingleWrapperState:
        return BaseSingleWrapperState(
            dataset_type=type(self).__name__,
            dataset_state=self.dataset.save_state(),
        )

    def merge_states(
        self, states: List[Optional[BaseSingleWrapperState]]
    ) -> BaseSingleWrapperMergedState:
        assert all(s is None or isinstance(s, BaseSingleWrapperState) for s in states)
        assert all(s is None or s.dataset_type == type(self).__name__ for s in states)
        return BaseSingleWrapperMergedState(
            dataset_type=type(self).__name__,
            dataset_state=self.dataset.merge_states(
                [None if s is None else s.dataset_state for s in states]
            ),
        )

    def restore_state(self, state: Optional[BaseSingleWrapperMergedState]) -> None:
        if state is None:
            self.dataset.restore_state(None)
        else:
            assert isinstance(state, BaseSingleWrapperMergedState)
            assert state.dataset_type == type(self).__name__
            self.dataset.restore_state(state.dataset_state)

    def can_restore_sample(self) -> bool:
        return self.dataset.can_restore_sample()

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_sample_out:
        return self.dataset.restore_sample(index)

    def verify_worker_config(self, worker_config: WorkerConfig) -> None:
        super().verify_worker_config(worker_config)
        self.dataset.verify_worker_config(worker_config)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Optional

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class ImageClassificationSample(Sample):
    """Sample type for classifying an image."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor
    #: The class label of the image
    label: Optional[int] = None
    #: The class label of the image
    label_name: Optional[str] = None


class ImageClassificationWebdataset(DefaultDecoderWebdataset[ImageClassificationSample]):
    __sample_type__ = ImageClassificationSample
    type = ImageClassificationSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, Dict, Generic, Iterator, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.wrappers.base import BaseSingleWrapperDataset

T_sample = TypeVar("T_sample")


class FilterDataset(BaseSingleWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """This dataset wrapper applies a custom filter function to each sample and does not yield
    filtered samples."""

    filter_fn: Callable[[T_sample], bool]

    def __init__(self, dataset: SavableDataset[T_sample], filter_fn: Callable[[T_sample], bool]):
        """Construct a MapDataset.

        Args:
            dataset: The input dataset to wrap
            filter_fn: The function to apply to each sample. If it returns `True`, the sample is
               accepted.
        """
        super().__init__(dataset)
        self.filter_fn = filter_fn

    def __len__(self):
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample]:
        for sample in self.dataset:
            if self.filter_fn(sample):
                yield sample

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "filter_fn": self._function_config(self.filter_fn),
        }

    def __str__(self):
        return f"FilterDataset(filter_fn={self.filter_fn}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path
from typing import Union

from megatron.energon.dataset_config import load_config
from megatron.energon.epathlib import EPath
from megatron.energon.metadataset.dataset_loader import DatasetLoader
from megatron.energon.metadataset.loader_interface import DatasetLoaderInterface
from megatron.energon.metadataset.metadataset import Metadataset


def load_dataset(
    path: Union[str, EPath, Path],
    **kwargs,
) -> DatasetLoaderInterface:
    """Loads a (meta)dataset."""
    path = EPath(path).absolute()

    if path.is_file():
        return load_config(
            path,
            default_type=Metadataset,
            strict=True,
            default_kwargs=dict(parent_path=path.parent, **kwargs),
        )
    else:
        return DatasetLoader(path=path, **kwargs)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, TypeVar

from torch.utils.data import DataLoader

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.savable_loader import BasicDataLoader, SavableDataLoader
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.gc_dataset import GcDataset

T = TypeVar("T")


def get_savable_loader(
    dataset: SavableDataset[T],
    *,
    worker_config: Optional[WorkerConfig] = None,
    checkpoint_every_sec: float = 60,
    checkpoint_every_min_n_samples: Optional[int] = None,
    n_checkpoints: int = 2,
) -> SavableDataLoader[T]:
    """

    Get a dataloader for the given dataset.

    Args:
        dataset: The dataset to create a loader for.
        worker_config: The worker config. Set the same value to the `get(_val|_train)_dataset` call.
        checkpoint_every_sec: This is the time in seconds after which an internal checkpoint is
            saved. It may take the same duration to restore a checkpoint, but introduces additional
            overhead during reading data from the dataset, so this should be chosen accordingly.
            Only applies if using workers.
        checkpoint_every_min_n_samples: Overwrites the minimum number of samples between
            checkpoints. Defaults to `number of workers * 2`. Only applies if using workers.
        n_checkpoints: The number of internal checkpoints to keep. Only applies if using workers.

    Returns:
        The instantiated :class:`megatron.energon.SavableDataLoader`, yielding batches from the dataset,
        allowing to save the state of the dataset.
    """
    if worker_config is None:
        worker_config = global_worker_config

    dataset.verify_worker_config(worker_config)
    return SavableDataLoader(
        dataset,
        worker_config=worker_config,
        checkpoint_every_sec=checkpoint_every_sec,
        checkpoint_every_min_n_samples=checkpoint_every_min_n_samples,
        n_checkpoints=n_checkpoints,
    )


def get_loader(
    dataset: SavableDataset[T],
    *,
    worker_config: Optional[WorkerConfig] = None,
) -> DataLoader[T]:
    """
    Get a dataloader for the given dataset.

    Args:
        dataset: The dataset to create a loader for.
        worker_config: The worker config. Set the same value to the `get(_val|_train)_dataset` call.

    Returns:
        The instantiated :class:`torch.data.DataLoader`, yielding batches from the dataset.
    """
    if worker_config is None:
        worker_config = global_worker_config

    dataset.verify_worker_config(worker_config)
    return BasicDataLoader(
        GcDataset(dataset),
        worker_config=worker_config,
    )



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import inspect
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass, is_dataclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import torch
from torch.utils.data import IterableDataset

from megatron.energon.epathlib import EPath
from megatron.energon.worker import WorkerConfig

T_sample = TypeVar("T_sample", covariant=True)
T = TypeVar("T", covariant=True)


class PinMemoryMixin:
    """A mixin class providing a generic `pin_memory` function."""

    def _pin_memory(self, batch: T, device: Union[torch.device, str, None] = None) -> T:
        """Pin memory of a batch. Uses recursion to handle nested structures. Supports nested
        structures of dicts, dataclasses, namedtuples, lists and tuples."""
        if isinstance(batch, torch.Tensor):
            return batch.pin_memory(device)
        elif isinstance(batch, dict):
            return {key: self._pin_memory(value, device) for key, value in batch.items()}
        elif dataclasses.is_dataclass(batch):
            return type(batch)(
                **{
                    field.name: self._pin_memory(getattr(batch, field.name), device)
                    for field in dataclasses.fields(batch)
                }
            )
        elif isinstance(batch, (tuple, list)):
            if hasattr(batch, "_fields"):
                # NamedTuple
                return type(batch)(*[self._pin_memory(val, device) for val in batch])
            else:
                # list / tuple
                return type(batch)(self._pin_memory(val, device) for val in batch)
        else:
            return batch

    def pin_memory(self: T) -> T:
        return self._pin_memory(self)


class ExtendableDataclassMixin:
    """A mixin class providing a generic `extend` function for copying dataclasses."""

    @classmethod
    def extend(cls: Type[T], src, **kwargs) -> T:
        """
        Used for overridden dataclass instances. Example

        .. code-block:: python

            @dataclass
            class MyBaseClass:
                a: List[int]

            @dataclass
            class MyExtendedClass(MyBaseClass):
                # Add a new field `b` to the state
                b: List[int]

            base = MyBaseClass(a=[1, 2, 3])
            extended = MyExtendedClass.extend(base, b=[4, 5, 6])

        Args:
            src: The source dataclass instance to extend.
            **kwargs: The new fields to add to the instance to construct the new instance.

        Returns:
            The extended dataclass instance.
        """
        assert is_dataclass(cls), "Must be a dataclass"
        assert issubclass(cls, type(src)), "Cannot extend class of different type"

        for f in dataclasses.fields(src):
            if not f.init or f.type is ClassVar or typing.get_origin(f.type) is ClassVar:
                continue

            if f.name not in kwargs:
                kwargs[f.name] = getattr(src, f.name)
        return cls(**kwargs)


@dataclass
class Sample(ABC, PinMemoryMixin, ExtendableDataclassMixin):
    """An abstract base class for one element of a batch.
    Each task should derive a specific subclass as a `@dataclass`, like
    :class:`megatron.energon.CaptioningBatchSample`, and add the input and output fields as needed for
    training.
    """

    #: Uniquely identifies each sample in the dataset.
    __key__: str
    #: Key for restoring the sample. This is used to restore the sample from a checkpoint. It
    # should be a tuple of strings and integers, which can be used to index the dataset.
    __restore_key__: Tuple[Union[str, int], ...]

    #: A dataset may define a subflavor to distinguish between samples of the same sample type.
    __subflavor__: Optional[str]
    #: A dataset may define a subflavors to distinguish between samples of the same sample type.
    __subflavors__: Optional[Dict[str, Any]]


@dataclass
class State(ABC, ExtendableDataclassMixin):
    """An abstract base class for the state of a dataset. See :class:`megatron.energon.SavableDataset`.
    The state of a dataset is used to save and restore the dataset state (i.e. random generators,
    buffer states, file pointers, etc.).
    Each dataset should derive a specific subclass as a `@dataclass` and add the fields as needed
    for training.

    To extend subclasses, use the .extend method. Example:

    .. code-block:: python

        @dataclass
        class MyState(State):
            a: int

        @dataclass
        class MyExtendedState(MyState):
            # Add a new field `b` to the state
            b: int

        class MyStateSaver:
            def save_state(self) -> MyState:
                return MyState(a=42)

        class MyExtendedStateSaver(MyStateSaver):
            def save_state(self) -> MyExtendedState:
                # Fetch state from super class, which is already a complete instance (cannot add
                # new fields to it, type is fixed).
                state: MyState = super().save_state()

                # Now extend the state of the super class (of type `MyState`) with the new field
                # required to define `MyExtendedState`.
                return MyExtendedState.extend(state, b=21)
    """


@dataclass
class MergedState(ABC, ExtendableDataclassMixin):
    """An abstract base class for the merged state of a dataset. See :class:`SavableDataset`.
    The merged state is created in the :meth:`megatron.energon.SavableDataset.merge_states` method, and
    represents the merged state of all worker processes (only workers, not ranks). It is required
    to restore the state of the dataset in the :meth:`megatron.energon.SavableDataset.restore_state`
    method for all workers before the workers are started.
    Each dataset should derive a specific subclass as a `@dataclass` and add the fields as needed
    for training.

    To extend subclasses, use the .extend method. Example:
    Example (see :meth:`megatron.energon.State` to complete the example)

    .. code-block:: python

        @dataclass
        class MyMergedState(MergedState):
            a: List[int]

        @dataclass
        class MyExtendedMergedState(MyMergedState):
            # Add a new field `b` to the state
            b: List[int]

        class MyMergedStateSaver:
            def merge_state(self, states: List[MyState]) -> MyMergedState:
                return MyMergedState(a=[s.a for s in states])

        class MyExtendedMergedStateSaver(MyMergedStateSaver):
            def merge_state(self, states: List[MyExtendedState]) -> MyExtendedMergedState:
                # Fetch state from super class, which is already a complete instance (cannot add
                # new fields to it, type is fixed).
                state: MyMergedState = super().merge_state(states)

                # Now extend the state of the super class (of type `MyMergedState`) with the
                # new field required to define `MyExtendedMergedState`.
                return MyExtendedMergedState.extend(state, b=[s.b for s in states])
    """


class SavableDataset(IterableDataset[T_sample], Generic[T_sample], ABC):
    """A dataset that can be saved and restored (i.e. the random state, internal buffers, etc.).
    I.e. it can be resumed from a checkpoint.

    How dataset state saving works:
    1. The dataset state needs to be saved in all forked worker processes which contain a copy of
      the main dataset instance (see :class:`megatron.energon.SavableDataLoader`). Each worker returns
      only its own state.
    2. The main process merges the states via the :meth:`megatron.energon.SavableDataset.merge_states`
      method in the main process on the main dataset instance (which doesn't hold the worker states,
      as they were forked).
    3. The main process saves the merged state to the checkpoint.
    """

    @abstractmethod
    def __len__(self) -> int: ...

    @abstractmethod
    def save_state(self) -> State:
        """
        Saves the state of the dataset. This should include the random state, but not the data
        itself. Can only be called in a worker process.

        Returns:
            The state of the dataset as savable object (i.e. python basic types).
        """
        ...

    @abstractmethod
    def merge_states(self, states: List[Optional[State]]) -> MergedState:
        """
        Merges the states of all workers into one state. Restore state expects a merged state.

        Args:
            states: The states of the workers to merge.

        Returns:
            The merged state.
        """
        ...

    @abstractmethod
    def restore_state(self, state: Optional[MergedState]) -> None:
        """
        Restores the state of the dataset. This should include the random state, but not the data
        itself. Can only be called in a worker process.

        Args:
            state: The state of the dataset as savable object (i.e. python basic types) as saved by
                `save_state`. If None, restore initial state.
        """
        ...

    def verify_worker_config(self, worker_config: WorkerConfig) -> None:
        """Verify that the worker config is compatible with the dataset."""
        if hasattr(self, "worker_config"):
            assert self.worker_config == worker_config, "Worker config is not consistent."

    @staticmethod
    def _function_config(fn: Callable) -> str:
        mod = inspect.getmodule(fn)
        if mod is not None:
            mod_name = mod.__name__
        else:
            mod_name = getattr(fn, "__module__", "<unknown>")
        return f"{mod_name}.{getattr(fn, '__qualname__', getattr(fn, '__name__', '<unknown>'))}"

    @abstractmethod
    def config(self) -> Dict[str, Any]:
        """Return a config dict that can be used to check if datasets have the same settings."""
        return {
            "type": type(self).__qualname__,
        }

    def can_restore_sample(self) -> bool:
        """Returns True if the dataset can restore a sample from a key."""
        return False

    def _add_sample_restore_key(
        self, sample: T_sample, key: Union[int, str], fail_otherwise: bool = False
    ) -> T_sample:
        """Adds a key to a sample. The sample must be a valid `Sample` or dict containing
        __restore_key__, which is a tuple of keys that can be used to restore the inner sample.
        This restore key is prepended with the `key`."""
        if isinstance(sample, Sample):
            sample.__restore_key__ = (key, *sample.__restore_key__)
        elif isinstance(sample, dict) and "__restore_key__" in sample:
            sample["__restore_key__"] = (key, *sample["__restore_key__"])
        elif fail_otherwise:
            raise RuntimeError(
                "Did not yield a sample with a restore key, but is marked "
                "stateless/deterministic."
            )
        return sample

    def restore_sample(self, index: Tuple[Union[str, int], ...]) -> T_sample:
        """
        Generic key type, because it might be either an integer (for a core dataset), or something
        more complex (e.g. for blended datasets).

        Default raises an exception (assumed non-deterministic if not implemented, does not
        guarantee determinism).
        """
        raise NotImplementedError(
            "This dataset does not support indexing, because it is not safely deterministic."
        )


class BaseCoreDataset(SavableDataset[T_sample], Generic[T_sample], ABC):
    """Base type for an inner dataset loaded from a .nv-meta folder."""

    __sample_type__: Type[T_sample] = cast(Type[T_sample], None)
    path: EPath

    subflavor: Optional[str]
    subflavors: Dict[str, Any]



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
import logging
import multiprocessing as mp
import random
import re
import tarfile
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Container,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import braceexpand
import webdataset
import yaml

from megatron.energon.dataset_config import MAIN_FOLDER_NAME
from megatron.energon.epathlib import EPath
from megatron.energon.errors import SYSTEM_EXCEPTIONS
from megatron.energon.flavors.base_dataset import (
    BaseCoreDataset,
    MergedState,
    Sample,
    SavableDataset,
    State,
)
from megatron.energon.flavors.webdataset.config import skip_meta_re, split_name_re
from megatron.energon.flavors.webdataset.field_access import field_access, split_field_access
from megatron.energon.flavors.webdataset.itar import TarIndexWriter
from megatron.energon.flavors.webdataset.sample_loader import WebdatasetSampleLoaderDataset
from megatron.energon.flavors.webdataset.structs import (
    FilteredSample,
    ShardInfo,
    WebdatasetInfo,
    WebdatasetSplits,
    reraise_exception,
)
from megatron.energon.module_loader import ModuleLoader
from megatron.energon.typed_converter import raw_to_typed, to_json_object
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.iter_map_dataset import IterMapDataset
from megatron.energon.wrappers.map_dataset import MapDataset

T_sample = TypeVar("T_sample", covariant=True)
T = TypeVar("T", covariant=True)

logger = logging.getLogger(__name__)

ImageDecoder = Literal[
    "l8",
    "rgb8",
    "rgba8",
    "l",
    "rgb",
    "rgba",
    "torchl8",
    "torchrgb8",
    "torchrgba8",
    "torchl",
    "torchrgb",
    "torch",
    "torchrgba",
    "pill",
    "pil",
    "pilrgb",
    "pilrgba",
]


@dataclass
class WebdatasetState(State):
    dataset_state: State


@dataclass
class WebdatasetMergedState(MergedState):
    dataset_state: MergedState


class BaseWebdataset(BaseCoreDataset[T_sample], Generic[T_sample], ABC):
    """
    Base class for all webdataset loaders. Applies proper sharding across workers.
    """

    path: EPath
    training: bool
    worker_config: WorkerConfig
    info: WebdatasetInfo
    split_config: WebdatasetSplits
    extension_fn: Optional[Callable[[str], bool]]

    shards: List[ShardInfo]
    dataset: SavableDataset[T_sample]
    rank_shards: List[List[ShardInfo]]

    class EmptyDatasetError(Exception):
        """Raised when a dataset is empty."""

    def __init__(
        self,
        path: EPath,
        *,
        split_part: str,
        training: bool,
        worker_config: Optional[WorkerConfig] = None,
        shuffle_over_epochs: int = 1,
        parallel_shard_iters: Optional[int] = None,
        max_samples_per_sequence: Optional[int] = None,
        info_config: str = ".info.yaml",
        split_config: str = "split.yaml",
        part_filter: Optional[Callable[[str], bool]] = None,
        handler: Callable[[Exception, Optional[str]], None] = reraise_exception,
    ):
        """
        Constructs the webdataset loader.

        Args:
            path: Root path to the dataset for relative path resolution.
            split_part: Which part to load (e.g. 'train', 'val', 'test').
            training: If true, apply shuffling and loop the dataset.
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
            shuffle_over_epochs: Only effective if training=True.
                How many epochs to shuffle over if training.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather shard slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the shards are effectively shuffle over infinite epochs (i.e. shard slices
                are drawn with replacement).
            parallel_shard_iters: Number of parallel opened shards per worker, shuffling between.
            max_samples_per_sequence: Maximum number of samples per sequence (=how many samples
                    will be sequentially iterated).
            info_config: Config file to use for sample metadata.
            split_config: Config file to use for shard split definitions.
            part_filter: (internal) Function for filtering tar files by dict keys
            handler: Exception handler. Args: (exception, key).
        """
        assert self.__sample_type__ is not None, f"Class {type(self)} must define __sample_type__"
        self.path = path
        self.training = training
        self.worker_config = worker_config or global_worker_config
        self.handler = handler
        self.info = raw_to_typed(
            yaml.safe_load((path / MAIN_FOLDER_NAME / info_config).read_text()), WebdatasetInfo
        )
        self.split_config = raw_to_typed(
            yaml.safe_load((path / MAIN_FOLDER_NAME / split_config).read_text()), WebdatasetSplits
        )
        assert split_part in self.split_config.split_parts, f"Invalid split part: {split_part!r}"
        self.shards = [
            ShardInfo(
                name=name,
                path=path / name,
                offset=0,
                count=self.info.shard_counts[name],
            )
            for name in self.split_config.split_parts[split_part]
            for name in braceexpand.braceexpand(name)
        ]
        if len(self.shards) == 0:
            raise BaseWebdataset.EmptyDatasetError(f"No shards found in split part {split_part!r}")

        exclude = {
            excluded
            for excluded in self.split_config.exclude
            for excluded in braceexpand.braceexpand(excluded)
        }
        total_shard_list = [shard for shard in self.shards if shard.name not in exclude]
        assert len(total_shard_list) > 0, f"No shards left after filtering."

        if parallel_shard_iters is None:
            if training:
                # 16 seems to be a good choice since we don't want too many file handles open
                parallel_shard_iters = 16
            else:
                parallel_shard_iters = 1

        self.rank_shards = self._shard_workers(
            total_shard_list,
            self.worker_config,
            max_samples_per_sequence=max_samples_per_sequence,
        )

        self.rank_total = sum(shard.count for shards in self.rank_shards for shard in shards)
        for idx, shards in enumerate(self.rank_shards):
            shards_text = ", ".join(
                f"{shard.name}[{shard.offset}, {shard.offset+shard.count})" for shard in shards[:3]
            )
            if len(shards) > 6:
                shards_text += f", ...<{len(shards) - 6}>, " + ", ".join(
                    f"{shard.name}[{shard.offset}, {shard.offset+shard.count})"
                    for shard in shards[-3:]
                )
            elif len(shards) > 3:
                shards_text += ", " + ", ".join(
                    f"{shard.name}[{shard.offset}, {shard.offset+shard.count})"
                    for shard in shards[3:]
                )
            print(
                f"rank={self.worker_config.rank}, worker={idx}: shard_range="
                f"[{shards_text}] "
                f"sum(count)={sum(s.count for s in shards)}"
            )

        dataset = WebdatasetSampleLoaderDataset(
            rank_shards=self.rank_shards,
            worker_config=self.worker_config,
            part_filter=part_filter,
            exclude=exclude,
            loop=training,
            shuffle_over_epochs=shuffle_over_epochs if training else None,
            parallel_shard_iters=parallel_shard_iters,
            handler=self.sample_error_handler,
        )
        self.dataset = self._process_samples(dataset)

    def sample_error_handler(self, e: Exception, sample_key: str):
        if isinstance(e, SYSTEM_EXCEPTIONS):
            raise SampleError(f"Error in sample {sample_key!r}: {e}") from e

        self.handler(e, sample_key)

    def error_handler(self, e: Exception, sample: Union[T_sample, List[T_sample], dict]):
        if isinstance(sample, dict):
            key = sample.get("__key__")
        elif isinstance(sample, list):
            if isinstance(sample[0], dict):
                key = ",".join(s.get("__key__") for s in sample)
            elif isinstance(sample[0], Sample):
                key = ",".join(s.__key__ for s in sample)
            else:
                key = None
        elif isinstance(sample, Sample):
            key = sample.__key__
        else:
            key = None
        self.sample_error_handler(e, key)

    @abstractmethod
    def _process_samples(self, dataset: SavableDataset[FilteredSample]) -> SavableDataset[T_sample]:
        """Internally loads the sample."""
        ...

    @staticmethod
    def _split_shard(
        shard: ShardInfo,
        start_offset: int,
        end_offset: int,
        max_samples_per_sequence: Optional[int],
    ) -> List[ShardInfo]:
        if (
            max_samples_per_sequence is not None
            and end_offset - start_offset > max_samples_per_sequence * 1.5
        ):
            # Split the shard into slices of max_samples_per_sequence (more or less)
            slice_count = max(round((end_offset - start_offset) / max_samples_per_sequence), 1)
            samples_per_sequence = (end_offset - start_offset) / slice_count
            # Note this must include the end offset as well, so slice_count + 1 steps (down there,
            # idx+1 is used to access the end offset)
            offsets = [
                start_offset + int(slice * samples_per_sequence) for slice in range(slice_count + 1)
            ]
            return [
                dataclasses.replace(
                    shard,
                    offset=offsets[idx],
                    count=offsets[idx + 1] - offsets[idx],
                    byte_offset=None,
                    byte_size=None,
                )
                for idx in range(slice_count)
            ]
        else:
            return [
                dataclasses.replace(
                    shard,
                    offset=start_offset,
                    count=end_offset - start_offset,
                    byte_offset=None,
                    byte_size=None,
                )
            ]

    @staticmethod
    def _split_shards(
        shards: List[ShardInfo], offsets: List[int], *, max_samples_per_sequence: Optional[int]
    ) -> Generator[List[ShardInfo], None, None]:
        """
        Splits the shards into multiple lists based on the offsets. The first offset is the start
        of the first shard emitted, the last offset is the end of the last shard emitted.
        (i.e. number of shards emitted is `len(offsets) - 1`)

        Args:
            shards: The source shards
            offsets: The offsets to samples to get shards for (must be strictly increasing)
            max_samples_per_sequence: Maximum number of samples per sequence (=how many samples
                  will be sequential).

        Returns:
            A list of shards for each offset pair
        """
        # The start index of the current shard
        cum_count = 0

        # Find shard idx for start
        for start_index, start_shard in enumerate(shards):
            if cum_count + start_shard.count < offsets[0]:
                # The shard is before the offset -> go to next shard
                cum_count += start_shard.count
                continue
            else:
                # The shard contains the offset
                start_offset = offsets[0] - cum_count
                break
        else:
            raise ValueError("Invalid shard distribution")

        for offset in offsets[1:]:
            # Find shard idx for end
            for end_index, end_shard in enumerate(shards[start_index:], start=start_index):
                if cum_count + end_shard.count < offset:
                    # The shard is before the offset -> go to next shard
                    cum_count += end_shard.count
                    continue
                else:
                    # The shard contains the offset
                    end_offset = offset - cum_count
                    break
            else:
                raise ValueError("Invalid shard distribution")
            if start_index == end_index:
                yield BaseWebdataset._split_shard(
                    start_shard,
                    start_offset=start_offset,
                    end_offset=end_offset,
                    max_samples_per_sequence=max_samples_per_sequence,
                )
            else:
                # Middle is the original shards, start and end get an offset/length
                yield (
                    (
                        BaseWebdataset._split_shard(
                            start_shard,
                            start_offset=start_offset,
                            end_offset=start_shard.count,
                            max_samples_per_sequence=max_samples_per_sequence,
                        )
                        if start_shard.count > start_offset
                        else []
                    )
                    + sum(
                        (
                            BaseWebdataset._split_shard(
                                shard,
                                start_offset=shard.offset,
                                end_offset=shard.count,
                                max_samples_per_sequence=max_samples_per_sequence,
                            )
                            for shard in shards[start_index + 1 : end_index]
                        ),
                        start=[],
                    )
                    + BaseWebdataset._split_shard(
                        end_shard,
                        start_offset=end_shard.offset,
                        end_offset=end_offset,
                        max_samples_per_sequence=max_samples_per_sequence,
                    )
                )
            start_index = end_index
            start_shard = end_shard
            start_offset = end_offset

    @classmethod
    def _shard_workers(
        cls,
        shards: List[ShardInfo],
        worker_config: WorkerConfig,
        *,
        max_samples_per_sequence: Optional[int],
    ) -> List[List[ShardInfo]]:
        """
        Creates subshards (ShardInfo) for each worker of the current rank.
        For that, the total number of samples is split into the number of global workers across all
        ranks. Then each worker gets a slice of the global samples.

        Args:
            shards: The shards to split
            worker_config: The config for the current rank and workers

        Returns:
            The shards for the current rank and all workers
        """

        # We split the total number of samples into the number of global workers across all ranks.
        # Note that the global number of workers intentionally stay the same if you
        # divide the number of ranks by N, and multiply the number of workers per rank by N.
        # This allows to reproduce the same global batches with a different number of ranks.
        
        num_workers = max(1, worker_config.num_workers)

        total_samples = sum(shard.count for shard in shards)
        global_workers = num_workers * worker_config.world_size

        samples_per_worker = total_samples / global_workers

        # Let's look at the workers of the current local rank.
        # Each worker gets a slice of the global samples as follows:
        local_rank_worker_sample_offsets = [
            int(
                (num_workers * worker_config.rank + local_worker_idx)
                * samples_per_worker
            )
            for local_worker_idx in range(num_workers + 1)
        ]

        return list(
            cls._split_shards(
                shards,
                local_rank_worker_sample_offsets,
                max_samples_per_sequence=max_samples_per_sequence,
            )
        )

    def __len__(self):
        # In the training case, the result is an approximation (i.e. number of different samples)
        return self.rank_total

    def __iter__(self) -> Iterator[T_sample]:
        yield from self.dataset

    def save_state(self) -> WebdatasetState:
        return WebdatasetState(
            dataset_state=self.dataset.save_state(),
        )

    def merge_states(self, states: List[WebdatasetState]) -> WebdatasetMergedState:
        assert all(s is None or isinstance(s, WebdatasetState) for s in states)
        return WebdatasetMergedState(
            dataset_state=self.dataset.merge_states(
                [None if s is None else s.dataset_state for s in states]
            ),
        )

    def restore_state(self, state: Optional[WebdatasetMergedState]) -> None:
        if state is None:
            self.dataset.restore_state(None)
        else:
            assert isinstance(state, WebdatasetMergedState)
            self.dataset.restore_state(state.dataset_state)

    @staticmethod
    def _preprocess_tar(
        path: Union[str, EPath], parent_path: Union[str, EPath], max_parts: int
    ) -> Tuple[ShardInfo, Set[str]]:
        """Process a single tar file, i.e. read the tarinfos, generate the tar index and return
        stats.

        Args:
            path: Path to the tar file.
            parent_path: Root path of the dataset.
            max_parts: Maximum number of different parts to return

        Returns:
            Tuple of shard info and found keys of the loaded dicts.
        """
        EPath.prepare_forked_process()  # Multiproc with fork

        path = EPath(path)
        shard_info = ShardInfo(name=path.relpath, path=path, offset=0, count=0)

        if not shard_info.path.is_absolute():
            parent_path = EPath(parent_path)
            assert parent_path.is_absolute(), f"Parent path must be absolute: {parent_path}"
            shard_info.path = parent_path / path

        try:
            # Note: Write to .tmp file first, then remove .tmp extension, to make sure only complete
            # files are used.
            with shard_info.path.open("rb") as f:
                with tarfile.open(fileobj=f, mode="r:*") as tar, TarIndexWriter(
                    shard_info.path
                ) as iw:
                    count = 0
                    parts = set()
                    last_base_name = None
                    for member in tar:
                        if not member.isreg():
                            continue
                        if member.name is None:
                            continue
                        if skip_meta_re.match(member.name):
                            continue

                        name_match = split_name_re.match(member.name)
                        if name_match is None:
                            continue

                        base_name = name_match.group(1)
                        if len(parts) < max_parts:
                            parts.add(name_match.group(2))

                        if last_base_name != base_name:
                            iw.append(member.offset)
                            last_base_name = base_name
                            count += 1
                    shard_info.count = count
                    iw.append(tar.offset)
            return shard_info, parts
        except BaseException:
            logger.exception(f"Shard failed to load: {path!r}. Skipping it.")
            return shard_info, set()

    @staticmethod
    def iter_dataset_content(
        path: Union[str, EPath],
        extract_keys: Container[str] = (),
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Yield example dataset content for a few samples.

        Args:
            path: Path to the tar file.
        """
        path = EPath(path)
        with path.open("rb") as f:
            with tarfile.open(fileobj=f, mode="r:*") as tar:
                last_base_name = None
                sample = {}
                for member in tar:
                    if not member.isreg():
                        continue
                    if member.name is None:
                        continue
                    if skip_meta_re.match(member.name):
                        continue

                    name_match = split_name_re.match(member.name)
                    if name_match is None:
                        continue

                    base_name = name_match.group(1)
                    if last_base_name != base_name:
                        if sample:
                            yield sample
                        sample = {}
                        last_base_name = base_name
                    if name_match:
                        if name_match.group(2) in extract_keys:
                            sample[name_match.group(2)] = tar.extractfile(member).read()
                        else:
                            sample[name_match.group(2)] = None
                if sample:
                    yield sample

    @staticmethod
    def prepare_dataset(
        parent_path: Union[Path, EPath],
        paths: List[str],
        *,
        split_parts_ratio: Optional[List[Tuple[str, float]]] = None,
        split_parts_patterns: Optional[List[Tuple[str, str]]] = None,
        info_config: str = ".info.yaml",
        split_config: str = "split.yaml",
        shuffle_seed: Optional[int] = 42,
        progress_fn: Callable[[List, int], Iterable] = (lambda x, l: x),
        workers: int = 32,
        tar_index_only: bool = False,
    ) -> Set[str]:
        """
        Preprocess the shards and write the split config. Preprocessing is done in parallel.
        Counts the number of samples in each shard.

        Args:
            parent_path: Common parent path for the shards
            paths: Paths to the shards
            split_parts_ratio: Names of splits and their ratio (will be normalized)
            split_parts_patterns: Names of splits and their path patterns
            info_config: Filename for the info config (`parent_path / '.nv-meta' / info_config`)
            split_config: Filename for the info config (`parent_path / '.nv-meta' / split_config`)
            shuffle_seed: Seed for shuffling shards before splitting into split_parts. None to
                disable.
            progress_fn: Callback for progress bar
            workers: Number of parallel workers for reading each shard
            tar_index_only: Only create tar-index, then exit

        Returns:
            The set of all parts found in the shards. But at most 50.
        """
        parent_path = EPath(parent_path).absolute()

        found_parts = set()
        paths = [path for path in paths for path in braceexpand.braceexpand(path)]
        shards = []

        assert parent_path.is_absolute(), f"Parent path must be absolute: {parent_path}"

        # use functools partial to pass parent_path to process_tar
        process_tar = functools.partial(
            BaseWebdataset._preprocess_tar,
            parent_path=parent_path.url,  # convert to url string, to avoid EPath in multiprocessing
            max_parts=50,
        )

        with mp.Pool(workers) as pool:
            for shard_info, cur_parts in progress_fn(pool.imap(process_tar, paths), len(paths)):
                if shard_info.count == 0:
                    # This shard failed to load. Skip it.
                    continue
                shards.append(shard_info)
                if len(found_parts) < 50:
                    found_parts.update(cur_parts)

        if tar_index_only:
            return found_parts

        (parent_path / MAIN_FOLDER_NAME).mkdir(exist_ok=True)

        # Save info
        info = WebdatasetInfo(
            shard_counts={shard.name: shard.count for shard in shards},
        )
        with (parent_path / MAIN_FOLDER_NAME / info_config).open("w") as wf:
            yaml.dump(to_json_object(info), wf)

        if split_parts_ratio is not None:
            # Normalize ratio
            total_ratio = sum(split_ratio for _, split_ratio in split_parts_ratio)
            split_parts_ratio = [
                (split_part, split_ratio / total_ratio)
                for split_part, split_ratio in split_parts_ratio
            ]
            # Sample from shards based on the split ratio from split parts
            split_shards = {}
            if shuffle_seed is not None:
                random.Random(shuffle_seed).shuffle(shards)
            split_total = 0
            split_offset = 0
            for split_part, split_ratio in split_parts_ratio:
                split_total += split_ratio
                split_end = int(len(shards) * split_total)
                split_shards[split_part] = [shard.name for shard in shards[split_offset:split_end]]
                split_offset = split_end
        else:
            assert (
                split_parts_patterns is not None
            ), "Require either split_parts_ratio or split_parts_patterns"
            # Sample from shards based on the split patterns from split parts
            split_shards = {}
            for split_part, split_pattern in split_parts_patterns:
                patterns = [
                    re.compile(pattern) for pattern in braceexpand.braceexpand(split_pattern)
                ]
                split_shards[split_part] = [
                    shard.name
                    for shard in shards
                    if any(pattern.match(shard.name) for pattern in patterns)
                ]
        # Save split config
        splits_config = WebdatasetSplits(split_parts=split_shards)
        with (parent_path / MAIN_FOLDER_NAME / split_config).open("w") as wf:
            yaml.dump(to_json_object(splits_config), wf)

        return found_parts

    def can_restore_sample(self) -> bool:
        return True

    def restore_sample(self, key: Tuple[Union[str, int], ...]) -> T_sample:
        return self.dataset.restore_sample(key)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "training": self.training,
            "path": str(self.path),
            "worker_config": self.worker_config.config(),
            "dataset": self.dataset.config(),
        }

    def __str__(self):
        return f"{type(self).__name__}(path={self.path}, dataset={self.dataset})"


class DefaultGenericWebdataset(BaseWebdataset[T_sample], Generic[T_sample]):
    """
    Default implementation of Webdataset for generic samples and the generic config interface.
    """

    _sample_loader: Callable[[Dict[str, Any]], Dict[str, Any]]

    def __init__(
        self,
        path: EPath,
        *,
        subflavor: Optional[str] = None,
        subflavors: Optional[Dict[str, Any]] = None,
        field_map: Optional[Dict[str, str]] = None,
        sample_loader: Optional[Union[str, Callable[[dict], dict]]] = None,
        part_filter: Optional[Union[str, List[str], Callable[[str], bool]]] = None,
        **kwargs,
    ):
        assert (field_map is None) != (
            sample_loader is None
        ), "Either field_map or sample_loader must be provided."
        if sample_loader is not None:
            assert (
                part_filter is not None
            ), "part_filter must be provided if sample_loader is provided."
            module_loader = ModuleLoader()
            if isinstance(sample_loader, str):
                sample_loader = module_loader.get_function(
                    sample_loader, "sample_loader", relative_path=path / MAIN_FOLDER_NAME
                )
            else:
                assert callable(sample_loader)
                sample_loader = sample_loader
            if isinstance(part_filter, list):
                parts = set(part_filter)
                part_filter = lambda part: part in parts
            elif isinstance(part_filter, str):
                part_filter = module_loader.get_function(
                    part_filter, "part_filter", relative_path=path / MAIN_FOLDER_NAME
                )
            else:
                assert callable(part_filter)
            self._sample_loader = sample_loader
        else:
            assert field_map is not None
            assert part_filter is None
            # Split field map fields by json[field][field]
            fields = {key: split_field_access(field) for key, field in field_map.items()}
            assert set(field.name for field in dataclasses.fields(self.__sample_type__)).issuperset(
                fields.keys()
            ) and set(
                field.name
                for field in dataclasses.fields(self.__sample_type__)
                if field.default is not dataclasses.MISSING
                and field.default_factory is not dataclasses.MISSING
            ).issubset(
                field_map.keys()
            ), f"field_map does not map to type {self.__sample_type__.__name__} fields"
            self._sample_loader = lambda sample: {
                k: field_access(sample, v) for k, v in fields.items()
            }
            parts = set(access[0] for options in fields.values() for access in options)
            part_filter = lambda part: part in parts
        inner_sample_loader = self._sample_loader
        self._sample_loader = lambda sample: {
            "__key__": sample["__key__"],
            **inner_sample_loader(sample),
            "__restore_key__": sample["__restore_key__"],
            "__subflavor__": self.subflavor,
            "__subflavors__": self.subflavors,
        }
        super().__init__(path, **kwargs, part_filter=part_filter)
        self.subflavor = subflavor
        self.subflavors = subflavors or {}

    def _process_samples(self, dataset: SavableDataset[FilteredSample]) -> SavableDataset[T_sample]:
        return MapDataset(
            dataset, self._load_sample, error_handler=self.error_handler, stateless_map_fn=True
        )

    def _load_sample(self, sample: FilteredSample) -> T_sample:
        return self.__sample_type__(**self._sample_loader(sample))

    def config(self) -> Dict[str, Any]:
        return {
            **super().config(),
            "subflavor": self.subflavor,
            "subflavors": self.subflavors,
            "sample_loader": self._function_config(self._sample_loader),
        }


class DefaultDecoderWebdataset(DefaultGenericWebdataset[T_sample], Generic[T_sample]):
    """
    Basic class for any dataset which contains images. Applies default wds loading logic for all
    known extensions.
    """

    #: Image decoding result type
    image_decode: ImageDecoder
    #: If true, ignore errors when decoding.
    ignore_decoder_errors: bool

    def __init__(
        self,
        path: EPath,
        *,
        image_decode: ImageDecoder = "torchrgb",
        ignore_decoder_errors: bool = False,
        **kwargs,
    ):
        """
        Decoder dataset.

        Args:
            path: Path to the dataset (passed to parent)
            image_decode: This defines the decoding results.
            ignore_decoder_errors: If true, ignore errors when decoding.
            **kwargs: Args passed to parent constructor
        """
        self.image_decode = image_decode
        self.ignore_decoder_errors = ignore_decoder_errors
        super().__init__(path, **kwargs)

    def _decode_error_handler(self, exc: Exception) -> bool:
        if self.ignore_decoder_errors:
            return True
        raise exc

    def _process_samples(self, dataset: SavableDataset[FilteredSample]) -> SavableDataset[T_sample]:
        decoder = webdataset.decode(self.image_decode, handler=self._decode_error_handler)
        dataset = IterMapDataset(
            dataset, decoder, error_handler=self.error_handler, stateless_iter_fn=True
        )
        return super()._process_samples(dataset)

    def config(self) -> Dict[str, Any]:
        return {
            **super().config(),
            "image_decode": self.image_decode,
            "ignore_decoder_errors": self.ignore_decoder_errors,
        }



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import tarfile
import warnings
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, Iterator, List, Optional, Set, Tuple

import torch

from megatron.energon.epathlib import EPath
from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import MergedState, SavableDataset, State
from megatron.energon.flavors.webdataset.config import skip_meta_re, split_name_re
from megatron.energon.flavors.webdataset.itar import get_itar_byte_offset, open_itar
from megatron.energon.flavors.webdataset.structs import FilteredSample, ShardInfo, reraise_exception
from megatron.energon.rng import WorkerRng, WorkerRngMergedState, WorkerRngState
from megatron.energon.worker import WorkerConfig, global_worker_config


@dataclass
class ShardState:
    """
    Represents the iteration state of a single shard. I.e. contains a reference to the shard and the
    reader's offset.
    """

    #: The shard of this state
    shard: ShardInfo
    #: Sample offset relative to the sub-shard file offset
    offset: int
    #: Byte offset relative to the sub-shard file byte offset
    byte_offset: int


@dataclass
class SampleLoaderState(State):
    """
    The savable state for the wds sample loader. Contains the active and pending shards.
    """

    #: Rng state
    rng: WorkerRngState
    #: Pending shards are the shards which have not yet been opened, but should be processed
    # in the current "epoch"
    pending_shards: Optional[List[ShardInfo]]
    #: The active shards are the currently opened shards. May contain `None`, if there are fewer
    # shards available (i.e. pending_shards empty) than parallel shard iterators requested.
    active_shards: Optional[List[Optional[ShardState]]]
    #: The total number of samples retrieved, it's just a monotonically increasing counter
    sample_count: int
    #: Number of epochs this dataset has been iterated over
    epoch_count: int
    #: Number of samples retrieved in current epoch
    epoch_sample_count: int


@dataclass
class SampleLoaderMergedState(MergedState):
    #: Rng state
    rng: WorkerRngMergedState
    #: Pending shards are the shards which have not yet been opened, but should be processed
    # in the current "epoch"
    pending_shards: List[Optional[List[ShardInfo]]]
    #: The active shards are the currently opened shards. May contain `None`, if there are fewer
    # shards available (i.e. pending_shards empty) than parallel shard iterators requested.
    active_shards: List[Optional[List[Optional[ShardState]]]]
    #: The total number of samples retrieved, it's just a monotonically increasing counter
    sample_count: List[int]
    #: Number of epochs this dataset has been iterated over
    epoch_count: List[int]
    #: The number of samples retrieved in current epoch
    epoch_sample_count: List[int]


class WebdatasetSampleLoaderDataset(SavableDataset[FilteredSample]):
    """Internal class for loading samples from webdataset shards"""

    #: All shards for all workers `shards[worker_idx][shard_idx]`
    shards: List[List[ShardInfo]]
    #: All shards for all workers accessible by name and offset
    # `shards[worker_idx][(shard_name, offset)]`, created lazily
    _shards_by_key: Optional[List[Dict[Tuple[str, int], ShardInfo]]] = None
    #: Paths to shards for all workers by name `shards[worker_idx][shard_name]`, created lazily
    _shard_paths_by_name: Optional[Dict[str, EPath]] = None
    #: The data parallel worker config
    worker_config: WorkerConfig
    # Sample keys to ignore
    exclude: Set[str]
    # File extensions to load (or None to load all)
    extensions: Optional[Set[str]]

    # If true, loop the shards
    loop: bool
    # If = 1, every sample is seen exactly once per epoch. If > 1, samples
    # (or rather shard slices) are shuffled within this number of epochs (i.e. randomly
    # selected without replacement). If None, the shards are effectively shuffle over
    # infinite epochs (i.e. shard slices are drawn with replacement).
    shuffle_over_epochs: Optional[int]
    # Number of parallel iterators to be opened simultaneously (and random sample between them)
    parallel_shard_iters: int
    # Error handler
    handler: Callable[[Exception, str], None]

    # Worker's random generator
    _worker_rng: WorkerRng
    #: Pending shards are the shards which have not yet been opened, but should be processed
    # in the current "epoch"
    _pending_shards: List[List[ShardInfo]]
    #: The active shards are the currently opened shards. May contain `None`, if there are fewer
    # shards available (i.e. pending_shards empty) than parallel shard iterators requested.
    _active_shards_state: List[Optional[List[Optional[ShardState]]]]
    #: The total number of samples retrieved, it's just a monotonically increasing counter
    _sample_count: List[int]
    #: Number of epochs this dataset has been iterated over
    _epoch_count: List[int]
    #: The number of samples retrieved in current epoch
    _epoch_sample_count: List[int]

    def __init__(
        self,
        rank_shards: List[List[ShardInfo]],
        worker_config: WorkerConfig,
        exclude: Set[str],
        part_filter: Optional[Callable[[str], bool]] = None,
        loop: bool = False,
        shuffle_over_epochs: Optional[int] = None,
        parallel_shard_iters: int = 1,
        handler: Callable[[Exception, str], None] = reraise_exception,
    ):
        """
        The webdataset loader. Iterates over the shard infos and yields the samples.

        Args:
            rank_shards: The shards to iterate over for each worker of the current rank.
            worker_config: The worker configuration.
            exclude: A set of strings of the form "<shard name>" or "<shard name>/<sample index>" to
                exclude from iteration.
            part_filter: If not None, use this function to filter out wds files.
            loop: If true, loop the shards indefinitely.
            shuffle_over_epochs: If None, disable shuffling.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather shard slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the shards are effectively shuffle over infinite epochs (i.e. shard slices
                are drawn with replacement).
            parallel_shard_iters: If > 1, samples are randomly drawn from parallel shard iterators.
                This will not impact performance, but increase randomness. If = 1, the shards are
                iterated in order.
            handler: Exception handler. Args: (exception, key).
        """
        super().__init__()
        self.shards = rank_shards
        self.worker_config = worker_config or global_worker_config
        self.exclude = exclude
        self.part_filter = part_filter
        self.loop = loop
        self.shuffle_over_epochs = shuffle_over_epochs
        self.parallel_shard_iters = parallel_shard_iters
        self.handler = handler
        self._worker_rng = WorkerRng(worker_config)
        self._pending_shards = [[] for _ in range(len(self.shards))]
        self._active_shards_state = [[None] * parallel_shard_iters for _ in range(len(self.shards))]
        self._sample_count = [0] * len(self.shards)
        self._epoch_count = [0] * len(self.shards)
        self._epoch_sample_count = [0] * len(self.shards)
        assert shuffle_over_epochs is None or shuffle_over_epochs == -1 or shuffle_over_epochs >= 1
        assert self.parallel_shard_iters >= 1

    @property
    def shards_by_key(self) -> List[Dict[Tuple[str, int], ShardInfo]]:
        if self._shards_by_key is None:
            self._shards_by_key = [
                {(shard.name, shard.offset): shard for shard in shards} for shards in self.shards
            ]
        return self._shards_by_key

    @property
    def shard_path_map(self) -> Dict[str, EPath]:
        if self._shard_paths_by_name is None:
            self._shard_paths_by_name = {
                shard.name: shard.path for shards in self.shards for shard in shards
            }
        return self._shard_paths_by_name

    def _shards_once(self, shards: List[ShardInfo]) -> List[ShardInfo]:
        """Possibly (re)shuffles the shards using the random generator."""
        if self.shuffle_over_epochs is None:
            # No shuffling
            return list(shards)
        elif self.shuffle_over_epochs == -1:
            # Shuffle with replacement (i.e. infinite epochs), effectively return as many shards
            # as are required for parallel shard iterators.
            # Next shards are drawn in the _shards_iter function.
            return [
                shards[self._worker_rng.randbelow(len(shards))]
                for _ in range(self.parallel_shard_iters)
            ]
        elif self.shuffle_over_epochs >= 1:
            # Shuffle without replacement (potentially over multiple epochs)
            return self._worker_rng.shuffle(shards * self.shuffle_over_epochs)
        else:
            raise ValueError(f"Invalid shuffle_over_epochs: {self.shuffle_over_epochs}")

    def _filter_files(self, fname: str, shard_name: str) -> bool:
        """Filter function for webdataset for excluding files from the shards."""
        if shard_name in self.exclude:
            return False

        # Get base_name and extension if available
        m = split_name_re.match(fname)
        if not m:
            return False
        base_name, ext = m.groups()

        if f"{shard_name}/{base_name}" in self.exclude:
            return False
        if self.part_filter is not None and not self.part_filter(ext):
            return False
        return True

    def _tarfile_sample_iter(
        self,
        tar_file: tarfile.TarFile,
        shard_info: ShardInfo,
        absolute_tar_begin_byte_offset: int,
    ) -> Generator[Tuple[FilteredSample, int], None, None]:
        group: Optional[FilteredSample] = None
        last_group_key: Optional[str] = None
        key: str = f"{shard_info.name}/UNKNOWN"

        for tarinfo in tar_file:
            try:
                fname = tarinfo.name
                if not tarinfo.isreg():
                    continue
                if fname is None:
                    continue
                if skip_meta_re.match(fname):
                    continue

                # Get base_name and extension if available
                m = split_name_re.match(fname)
                if not m:
                    continue
                base_name, ext = m.groups()

                key = f"{shard_info.name}/{base_name}"

                if last_group_key != base_name:
                    if group is not None:
                        yield group, tarinfo.offset

                    group = dict(
                        __key__=key,
                        __shard__=shard_info.name,
                        __restore_key__=(
                            shard_info.name,
                            tarinfo.offset + absolute_tar_begin_byte_offset,
                        ),
                    )
                    last_group_key = base_name

                if self.part_filter is None or self.part_filter(ext):
                    group[ext] = tar_file.extractfile(tarinfo).read()
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample_key(key)
            except Exception as e:
                self.handler(e, key)

        if group is not None:
            # shard_state.byte_offset = (absolute_tar_begin_offset - shard_info.byte_offset)
            yield group, shard_info.byte_size - (
                absolute_tar_begin_byte_offset - shard_info.byte_offset
            )
        else:
            return

    def _shard_iter(self, shard_state: ShardState) -> Generator[FilteredSample, None, None]:
        """Iterates the samples in a shard (potentially resuming from a saved state)."""
        # print(
        #     f"Shard iter for shard {shard_state.shard.name} [{shard_state.shard.offset} +{shard_state.byte_offset}b, {shard_state.shard.offset + shard_state.shard.count} -{shard_state.byte_offset}b) starting"
        # )
        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "WebdatasetSampleLoaderDataset._shard_iter",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "shard": {
                        "name": shard_state.shard.name,
                        "path": str(shard_state.shard.path),
                        "offset": shard_state.shard.offset,
                        "count": shard_state.shard.count,
                    },
                    "offset": shard_state.offset,
                }
            )

        shard = shard_state.shard
        if shard.byte_offset is None:
            shard.byte_offset = get_itar_byte_offset(shard.path, shard.offset)
        if shard.byte_size is None:
            shard.byte_size = (
                get_itar_byte_offset(shard.path, shard.offset + shard.count) - shard.byte_offset
            )

        if shard_state.byte_offset == shard.byte_size:
            # Empty shard, return immediately (cannot be handled by the open_itar function)
            return

        # If the shard is not empty, the absolute byte offset must be smaller than the shard size
        assert shard_state.byte_offset < shard.byte_size

        # Given the shard offset (e.g. sub-shard) and the relative byte offset from the stored state, compute the absolute byte offset
        absolute_byte_offset = shard.byte_offset + shard_state.byte_offset
        sub_tar_byte_size = shard.byte_size - shard_state.byte_offset
        orig_shard_state_byte_offset = shard_state.byte_offset

        try:
            # Open the shard and resume where it was last stopped

            with open_itar(
                shard.path,
                byte_offset=absolute_byte_offset,
                byte_size=sub_tar_byte_size,
            ) as it:
                for group, next_sample_offset_in_sub_tar in self._tarfile_sample_iter(
                    it, shard, absolute_byte_offset
                ):
                    if self.worker_config.should_log(level=3):
                        self.worker_config.worker_log(
                            {
                                "t": "WebdatasetSampleLoaderDataset._shard_iter.yield",
                                "r": self.worker_config.rank,
                                "w": self.worker_config.rank_worker_id(),
                                "shard": {
                                    "name": shard_state.shard.name,
                                    "path": str(shard_state.shard.path),
                                    "offset": shard_state.shard.offset,
                                    "count": shard_state.shard.count,
                                },
                                "offset": shard_state.offset,
                                "key": group["__key__"],
                            }
                        )

                    # Next state
                    # NOTE: The next_sample_offset_in_sub_tar (tarinfo.offset) is relative to
                    # absolute_byte_offset, since open_itar crops a part out of the file.
                    # But we want to compute the offset relative to shard.byte_offset
                    shard_state.byte_offset = (
                        next_sample_offset_in_sub_tar + orig_shard_state_byte_offset
                    )
                    assert shard_state.byte_offset <= shard.byte_size
                    # print(f"Yield {group['__key__']} @next: {key} @{shard_state.byte_offset}b")

                    shard_state.offset += 1
                    if group["__key__"] in self.exclude:
                        continue
                    yield group

                # Set to end of subtar
                if shard_state.offset != shard.count or shard_state.byte_offset != shard.byte_size:
                    warnings.warn(
                        f"shard_state.offset({shard_state.offset}) != shard.count({shard.count}) or "
                        f"shard_state.byte_offset({shard_state.byte_offset}) != shard.byte_size({shard.byte_size})"
                        f"; this indicates an internal bug. Shard might not have been interated completely, samples may be missing."
                    )
                # assert shard_state.offset == shard.count
                # assert shard_state.byte_offset == shard.byte_size
        except SYSTEM_EXCEPTIONS:
            raise FatalSampleError.from_sample_key(f"{shard.path}")
        except Exception as e:
            self.handler(e, shard.name)
        # print(f"Shard iter for shard {shard.name} [{shard.offset}, {shard.count}) done")

    def _shards_iter(self, shards: List[ShardInfo]) -> Generator[FilteredSample, None, None]:
        """Iterates the samples in a list of shards, possibly looping over them indefinitely,
        possibly reshuffling the shards every loop, possibly using multiple parallel iterators over
        the shards."""
        worker_idx = self.worker_config.rank_worker_id()
        # Cleanup other states
        for i in range(self.worker_config.num_workers):
            if i != worker_idx:
                self._active_shards_state[i] = [None] * self.parallel_shard_iters
                self._pending_shards[i] = []
        shards_probs = torch.empty(self.parallel_shard_iters, dtype=torch.float32)
        shard_iters: List[Optional[Generator[FilteredSample, None, None]]]
        active_shards: List[Optional[ShardState]]
        while True:
            shards_probs[:] = 0
            shard_iters = []
            if (
                any(s is not None for s in self._active_shards_state[worker_idx])
                or len(self._pending_shards[worker_idx]) > 0
            ):
                shards_order = self._pending_shards[worker_idx]

                # Restore the state
                active_shards = self._active_shards_state[worker_idx]
                assert len(active_shards) == self.parallel_shard_iters
                for shard_state in active_shards:
                    if shard_state is None:
                        shard_iters.append(None)
                    else:
                        shards_probs[len(shard_iters)] = shard_state.shard.count
                        shard_iters.append(self._shard_iter(shard_state))

                if self.worker_config.should_log(level=1):
                    self.worker_config.worker_log(
                        {
                            "t": "WebdatasetSampleLoaderDataset._shards_iter.resume_epoch",
                            "r": self.worker_config.rank,
                            "w": self.worker_config.rank_worker_id(),
                            "shards": [
                                {
                                    "name": shard.name,
                                    "path": str(shard.path),
                                    "offset": shard.offset,
                                    "count": shard.count,
                                }
                                for shard in shards_order
                            ],
                            "active_shards": [
                                (
                                    None
                                    if shard_state is None
                                    else {
                                        "shard": {
                                            "name": shard_state.shard.name,
                                            "path": str(shard_state.shard.path),
                                            "offset": shard_state.shard.offset,
                                            "count": shard_state.shard.count,
                                        },
                                        "offset": shard_state.offset,
                                    }
                                )
                                for shard_state in active_shards
                            ],
                            "count": self._sample_count[worker_idx],
                            "epoch": self._epoch_count[worker_idx],
                            "epoch_count": self._epoch_sample_count[worker_idx],
                            "probs": shards_probs.tolist(),
                        }
                    )

            else:
                # Weight the shards by their size to get a more even distribution of samples
                shards_order = self._shards_once(shards)
                shards_order.reverse()

                if self.worker_config.should_log(level=1):
                    self.worker_config.worker_log(
                        {
                            "t": "WebdatasetSampleLoaderDataset._shards_iter.next_epoch",
                            "r": self.worker_config.rank,
                            "w": self.worker_config.rank_worker_id(),
                            "shards": [
                                {
                                    "name": shard.name,
                                    "path": str(shard.path),
                                    "offset": shard.offset,
                                    "count": shard.count,
                                }
                                for shard in shards_order
                            ],
                            "count": self._sample_count[worker_idx],
                            "epoch": self._epoch_count[worker_idx],
                            "epoch_count": self._epoch_sample_count[worker_idx],
                            "probs": shards_probs.tolist(),
                            "shuffle_over_epochs": self.shuffle_over_epochs,
                        }
                    )

                # List of shard iterators, always of length `parallel_shard_iters`. May contain `None`.
                active_shards = []
                # Fill up the shard iterators
                while len(shards_order) > 0 and len(shard_iters) < self.parallel_shard_iters:
                    shard = shards_order.pop()
                    shard_state = ShardState(shard=shard, byte_offset=0, offset=0)
                    shards_probs[len(shard_iters)] = shard.count
                    shard_iters.append(self._shard_iter(shard_state))
                    active_shards.append(shard_state)
                # Fill up the shard iterators with None
                for _ in range(len(shard_iters), self.parallel_shard_iters):
                    shard_iters.append(None)
                    active_shards.append(None)

                self._active_shards_state[worker_idx] = active_shards
                self._pending_shards[worker_idx] = shards_order

            # print(
            #     f"Next shard iters generated for {self.worker_config.rank}:{self.worker_config.rank_worker_id()}: probs={shards_probs}"
            # )

            # Iterate over the shard iterators
            while True:
                if torch.count_nonzero(shards_probs).item() == 0:
                    # There is no iterator left
                    break
                if self.shuffle_over_epochs is None:
                    # No shuffling, deterministic order, always the same
                    assert self.parallel_shard_iters == 1
                    shard_iter = shard_iters[0]
                else:
                    # Take a random shard iterator
                    shard_iter = self._worker_rng.choice(shard_iters, probs=shards_probs)
                try:
                    sample: FilteredSample = next(shard_iter)
                except StopIteration:
                    # Iterator exhausted -> take next / remove from list
                    rm_idx = shard_iters.index(shard_iter)
                    if len(shards_order) > 0 or self.shuffle_over_epochs == -1:
                        if len(shards_order) > 0:
                            # Take the next shard (without replacement)
                            shard = shards_order.pop()
                        else:
                            # Randomly select a new shard directly (with replacement)
                            shard = self.shards[worker_idx][
                                self._worker_rng.randbelow(len(self.shards[worker_idx]))
                            ]
                        shard_state = ShardState(shard=shard, byte_offset=0, offset=0)
                        shard_iters[rm_idx] = self._shard_iter(shard_state)
                        shards_probs[rm_idx] = shard.count
                        active_shards[rm_idx] = shard_state
                        # print(
                        #     f"Shard iter for {self.worker_config.rank}:{self.worker_config.rank_worker_id()} exhausted, taking next shard {shard.name} [{shard.offset}, {shard.offset + shard.count}), {len(shards_order)} shards left, probs={shards_probs}"
                        # )
                    else:
                        shard_iters[rm_idx] = None
                        shards_probs[rm_idx] = 0
                        active_shards[rm_idx] = None
                        # print(
                        #     f"Shard iter for {self.worker_config.rank}:{self.worker_config.rank_worker_id()} exhausted, no next shards, probs={shards_probs}"
                        # )
                    if self.worker_config.should_log(level=2):
                        self.worker_config.worker_log(
                            {
                                "t": "WebdatasetSampleLoaderDataset._shards_iter.exhausted",
                                "r": self.worker_config.rank,
                                "w": self.worker_config.rank_worker_id(),
                                "remaining": len(shards_order),
                                "count": self._sample_count[worker_idx],
                                "epoch": self._epoch_count[worker_idx],
                                "epoch_count": self._epoch_sample_count[worker_idx],
                                "probs": shards_probs.tolist(),
                            }
                        )
                else:
                    self._sample_count[worker_idx] += 1
                    self._epoch_sample_count[worker_idx] += 1
                    if self.worker_config.should_log(level=1):
                        self.worker_config.worker_log(
                            {
                                "t": "WebdatasetSampleLoaderDataset._shards_iter.yield",
                                "r": self.worker_config.rank,
                                "w": self.worker_config.rank_worker_id(),
                                "key": sample["__key__"],
                                "shard": sample["__shard__"],
                                "count": self._sample_count[worker_idx],
                                "epoch": self._epoch_count[worker_idx],
                                "epoch_count": self._epoch_sample_count[worker_idx],
                            }
                        )
                    yield sample
            if self.worker_config.should_log(level=2):
                self.worker_config.worker_log(
                    {
                        "t": "WebdatasetSampleLoaderDataset._shards_iter.all_exhausted",
                        "r": self.worker_config.rank,
                        "w": self.worker_config.rank_worker_id(),
                        "count": self._sample_count[worker_idx],
                        "epoch": self._epoch_count[worker_idx],
                        "epoch_count": self._epoch_sample_count[worker_idx],
                    }
                )

            self._epoch_count[worker_idx] += 1
            self._epoch_sample_count[worker_idx] = 0
            # print(
            #     f"Shard iters exhausted for {self.worker_config.rank}:{self.worker_config.rank_worker_id()} after {cnt} samples"
            # )
            if not self.loop:
                break

    def __len__(self) -> int:
        return sum(shard.count for shard in self.shards)

    def __iter__(self) -> Iterator[FilteredSample]:
        self.worker_config.assert_worker()
        worker_shards = self.shards[self.worker_config.rank_worker_id()]

        if self.worker_config.should_log(level=1):
            self.worker_config.worker_log(
                {
                    "t": "WebdatasetSampleLoaderDataset.__iter__",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "shard_range": [
                        f"{shard.name}[{shard.offset}, {shard.offset+shard.count})"
                        for shard in worker_shards
                    ],
                    "parallel_shard_iters": self.parallel_shard_iters,
                    "shuffle_over_epochs": self.shuffle_over_epochs,
                    "loop": self.loop,
                }
            )

        if len(worker_shards) == 0:
            return

        yield from self._shards_iter(worker_shards)

    def can_restore_sample(self) -> bool:
        return True

    def restore_sample(self, key: Tuple[str, int]) -> FilteredSample:
        shard_name, tar_byte_offset = key
        shard_path = self.shard_path_map[shard_name]

        sample_shard_info = ShardInfo(
            name=shard_name,
            path=shard_path,
            offset=0,
            count=1,
            byte_offset=tar_byte_offset,
            # Just a dummy value
            byte_size=0,
        )

        # Open the shard and extract the sample
        with open_itar(shard_path, byte_offset=tar_byte_offset) as tar_file:
            gen = self._tarfile_sample_iter(tar_file, sample_shard_info, tar_byte_offset)
            sample, _ = next(gen)
            gen.close()
            return sample

    def save_state(self) -> SampleLoaderState:
        self.worker_config.assert_worker()
        worker_idx = self.worker_config.rank_worker_id()
        assert len(self._active_shards_state[worker_idx]) == self.parallel_shard_iters
        if self.worker_config.should_log(level=3):
            self.worker_config.worker_log(
                {
                    "t": "WebdatasetSampleLoaderDataset.save_state",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "count": self._sample_count[worker_idx],
                    "epoch": self._epoch_count[worker_idx],
                    "epoch_count": self._epoch_sample_count[worker_idx],
                    "pending_shards": [
                        {
                            "name": shard.name,
                            "path": str(shard.path),
                            "offset": shard.offset,
                            "count": shard.count,
                        }
                        for shard in self._pending_shards[worker_idx]
                    ],
                    "active_shards": [
                        (
                            None
                            if shard_state is None
                            else {
                                "shard": {
                                    "name": shard_state.shard.name,
                                    "path": str(shard_state.shard.path),
                                    "offset": shard_state.shard.offset,
                                    "count": shard_state.shard.count,
                                },
                                "offset": shard_state.offset,
                            }
                        )
                        for shard_state in self._active_shards_state[worker_idx]
                    ],
                }
            )
        return SampleLoaderState(
            rng=self._worker_rng.save_state(),
            pending_shards=list(self._pending_shards[worker_idx]),
            active_shards=[
                None if active_shard is None else dataclasses.replace(active_shard)
                for active_shard in self._active_shards_state[worker_idx]
            ],
            sample_count=self._sample_count[worker_idx],
            epoch_count=self._epoch_count[worker_idx],
            epoch_sample_count=self._epoch_sample_count[worker_idx],
        )

    def merge_states(self, states: List[SampleLoaderState]) -> SampleLoaderMergedState:
        assert all(s is None or isinstance(s, SampleLoaderState) for s in states)
        assert len(states) == len(self.shards)
        return SampleLoaderMergedState(
            rng=self._worker_rng.merge_states([None if s is None else s.rng for s in states]),
            pending_shards=[[] if s is None else s.pending_shards for s in states],
            active_shards=[
                [None] * self.parallel_shard_iters if s is None else s.active_shards for s in states
            ],
            sample_count=[0 if s is None else s.sample_count for s in states],
            epoch_count=[0 if s is None else s.epoch_count for s in states],
            epoch_sample_count=[0 if s is None else s.epoch_sample_count for s in states],
        )

    def _restore_find_shard(
        self, shard_data: ShardInfo, shards_by_key: Dict[Tuple[str, int], ShardInfo]
    ) -> ShardInfo:
        shard = shards_by_key[(shard_data.name, shard_data.offset)]
        if shard != shard_data:
            raise ValueError(
                f"Shard {shard_data!r} not found in {self.shards!r}, states differ, not recoverable"
            )
        # Copy over the byte size and offset. Saves some loading time,
        # especially for restoring lots of random samples :)
        if shard_data.byte_offset is not None and shard.byte_offset is None:
            shard.byte_offset = shard_data.byte_offset
        if shard_data.byte_size is not None and shard.byte_size is None:
            shard.byte_size = shard_data.byte_size
        return shard

    def restore_state(self, state: Optional[SampleLoaderMergedState]) -> None:
        if self.worker_config.should_log(level=3):
            self.worker_config.worker_log(
                {
                    "t": "WebdatasetSampleLoaderDataset.restore_state",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "state": str(state),
                }
            )
        # print(f"Restore state {state}")
        if state is None:
            # Restore initial state
            self._worker_rng.restore_state(None)
            self._pending_shards = [[] for _ in range(len(self.shards))]
            self._active_shards_state = [
                [None] * self.parallel_shard_iters for _ in range(len(self.shards))
            ]
            self._sample_count = [0] * len(self.shards)
            self._epoch_count = [0] * len(self.shards)
            self._epoch_sample_count = [0] * len(self.shards)
        else:
            assert isinstance(state, SampleLoaderMergedState)
            self._worker_rng.restore_state(state.rng)
            # Restore state
            assert len(state.pending_shards) == len(self.shards)
            assert len(state.active_shards) == len(self.shards)
            self._pending_shards = [
                [self._restore_find_shard(shard, shards_by_key) for shard in pending_shards]
                for pending_shards, shards_by_key in zip(state.pending_shards, self.shards_by_key)
            ]
            self._active_shards_state = [
                [
                    (
                        None
                        if shard_state is None
                        else ShardState(
                            shard=self._restore_find_shard(shard_state.shard, shards_by_key),
                            offset=shard_state.offset,
                            byte_offset=shard_state.byte_offset,
                        )
                    )
                    for shard_state in active_shard_states
                ]
                for active_shard_states, shards_by_key in zip(
                    state.active_shards, self.shards_by_key
                )
            ]
            self._sample_count = state.sample_count
            self._epoch_count = state.epoch_count
            self._epoch_sample_count = state.epoch_sample_count

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "shards": [
                [
                    {
                        "name": shard.name,
                        "path": str(shard.path),
                        "offset": shard.offset,
                        "count": shard.count,
                        "byte_offset": shard.byte_offset,
                        "byte_size": shard.byte_size,
                    }
                    for shard in shards
                ]
                for shards in self.shards
            ],
            "worker_config": self.worker_config.config(),
            "exclude": list(self.exclude),
            "loop": self.loop,
            "shuffle_over_epochs": self.shuffle_over_epochs,
            "parallel_shard_iters": self.parallel_shard_iters,
        }

    def __str__(self):
        return f"WebdatasetSampleLoaderDataset(shards={self.shards}, loop={self.loop}, shuffle_over_epochs={self.shuffle_over_epochs}, parallel_shard_iters={self.parallel_shard_iters})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

"""Main entry point for energon CLI tools."""

import click

from megatron.energon.tools.analyze_debug import command as analyze_debug_command
from megatron.energon.tools.lint import command as lint_command
from megatron.energon.tools.prepare import command as prepare_command
from megatron.energon.tools.preview import command as preview_command


@click.group(context_settings=dict(help_option_names=["-h", "--help"]), invoke_without_command=True)
@click.pass_context
def main(ctx):
    """A set of tools that energon provides.

    Among other things, you can use it to lint or preprocess your dataset.

    See help of commands to learn more."""

    # This is needed to show help if no subcommand is provided
    if ctx.invoked_subcommand is None:
        click.echo(main.get_help(ctx))


main.add_command(analyze_debug_command)
main.add_command(lint_command)
main.add_command(prepare_command)
main.add_command(preview_command)

if __name__ == "__main__":
    main()



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, Generator, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.base import (
    BaseSingleWrapperDataset,
    BaseSingleWrapperMergedState,
    BaseSingleWrapperState,
)

T_sample = TypeVar("T_sample")


@dataclass
class EpochizeState(BaseSingleWrapperState):
    offset: int


@dataclass
class EpochizeMergedState(BaseSingleWrapperMergedState):
    offset: List[int]


class EpochizeDataset(BaseSingleWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """
    Uses the base dataset, and creates one epoch, which has length samples. Keeps the underlying
    dataset iterator alive over epochs (i.e. if it is an infinite dataset, it will keep the state).
    Repeats the underlying dataset if the iterator is exhausted.
    """

    length: int
    worker_config: WorkerConfig
    _active_iter: Optional[Iterator[T_sample]]

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        length: int,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """
        Create the epochized dataset.

        Args:
            dataset: The source dataset (possibly infinite)
            length: Number of samples to iterate before iteration stops (i.e. one epoch). When
                iteration continues, the original dataset iterator is resumed and does only restart
                if exhausted.
            worker_config: Configuration for the workers. Defaults to `global_worker_config`.
        """
        super().__init__(dataset)
        self.length = length
        self.worker_config = worker_config or global_worker_config
        self._offset = [0] * max(self.worker_config.num_workers, 1)
        self._active_iter = None

    def _infinite(self) -> Generator[T_sample, None, None]:
        while True:
            for sample in self.dataset:
                yield sample

    def __iter__(self) -> Iterator[T_sample]:
        # Compute the local length for this worker, i.e. all worker's lengths sum up to the total
        worker_idx = self.worker_config.rank_worker_id()

        if self.worker_config.num_workers <= 1:
            local_length = self.length
        else:
            local_length = self.length // self.worker_config.num_workers
            if self.worker_config.rank_worker_id() < self.length % self.worker_config.num_workers:
                local_length += 1

        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "EpochizeDataset.epoch_start",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "offset": self._offset[worker_idx],
                    "local_length": local_length,
                    "length": self.length,
                }
            )

        offset_range = list(range(self._offset[worker_idx], local_length))

        # Only iterate if there are samples to iterate
        if len(offset_range) > 0:
            if self._active_iter is None:
                self._active_iter = iter(self._infinite())

            for idx in offset_range:
                self._offset[worker_idx] = (idx + 1) % local_length
                yield next(self._active_iter)

        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "EpochizeDataset.epoch_end",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "offset": self._offset[worker_idx],
                    "local_length": local_length,
                    "length": self.length,
                }
            )

    def __len__(self) -> int:
        return self.length

    def save_state(self) -> EpochizeState:
        return EpochizeState.extend(
            super().save_state(), offset=self._offset[self.worker_config.rank_worker_id()]
        )

    def merge_states(self, states: List[EpochizeState]) -> EpochizeMergedState:
        assert all(s is None or isinstance(s, EpochizeState) for s in states)
        return EpochizeMergedState.extend(
            super().merge_states(states),
            offset=[0 if state is None else state.offset for state in states],
        )

    def restore_state(self, state: Optional[EpochizeMergedState]) -> None:
        super().restore_state(state)
        if state is None:
            self._offset = [0] * max(self.worker_config.num_workers, 1)
        else:
            assert isinstance(state, EpochizeMergedState)
            self._offset = state.offset

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "length": self.length,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"EpochizeDataset(length={self.length}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, TypedDict

from megatron.energon.epathlib import EPath


@dataclass
class WebdatasetInfo:
    # Maps shard name to number of samples in that shard
    shard_counts: Dict[str, int]


@dataclass
class WebdatasetSplits:
    # Maps split part to list of shard names
    split_parts: Dict[str, List[str]]
    # Set of "<shard name>" or "<shard name>/<sample index>" to exclude
    exclude: List[str] = field(default_factory=list)


@dataclass
class ShardInfo:
    #: Name of the shard file (relative path from the nvinfo dir)
    name: str
    #: The path to the shard file
    path: EPath
    #: The offset of the first sample (counted in samples) where this sub shard starts
    offset: int
    #: The number of samples in this sub shard
    count: int
    #: The absolute file offset in bytes where this sub shard starts, None if not computed yet
    byte_offset: Optional[int] = field(default=None, compare=False)
    #: The length of the sub shard in bytes within the shard file, None if not computed yet
    byte_size: Optional[int] = field(default=None, compare=False)


class FilteredSample(TypedDict):
    #: The key of the sample within the tar file.
    #: If the tar file contains files 12.jpg and 12.txt,
    #: those two files make one sample with the key "12"
    __key__: str
    #: The base name of the shard file e.g. "shard_000"
    __shard__: str
    #: Globally unique key to restore a sample from disk.
    #: For example `("shard_000", 123)` would restore
    #: the sample from "shard_000.tar" at byte offset 123.
    __restore_key__: Tuple[str, int]


def reraise_exception(exc: Exception, key: Optional[str]) -> None:
    if key:
        raise Exception(f"For sample {key!r}") from exc
    else:
        raise



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path
from typing import Any, Dict, Optional, Type, TypeVar, Union

import yaml

from megatron.energon.epathlib import EPath
from megatron.energon.flavors import BaseCoreDataset, Sample
from megatron.energon.typed_converter import raw_to_instance
from megatron.energon.worker import WorkerConfig

T = TypeVar("T")

MAIN_FOLDER_NAME = ".nv-meta"


def load_config(
    path: EPath,
    *,
    default_type: Optional[Type[T]] = None,
    strict: bool = True,
    default_kwargs: Optional[Dict[str, Any]] = None,
) -> T:
    """
    Loads a config from a file.

    Args:
        path: Path to the config to load
        default_type: If set, this is the type to use if no type is specified in the config.
        strict: If true, don't allow additional attributes in the config.
        default_kwargs: Default kwargs to use, will be overridden by the config.

    Returns:
        The instantiated type.
    """
    # Read the config
    path = path.absolute()
    with path.open() as f:
        data: dict = yaml.safe_load(f)
    if default_kwargs is not None:
        new_data = default_kwargs.copy()
        new_data.update(data)
        data = new_data
    return raw_to_instance(data, default_type, strict=strict)


T_sample = TypeVar("T_sample", covariant=True)


def get_dataset_from_config(
    path: Union[EPath, Path, str],
    *,
    dataset_config: str = "dataset.yaml",
    split_config: str = "split.yaml",
    split_part: str = "train",
    training: bool = True,
    subflavor: Optional[str] = None,
    subflavors: Optional[Dict[str, Any]] = None,
    worker_config: Optional[WorkerConfig] = None,
    sample_type: Optional[Type[T_sample]] = None,
    **kwargs,
) -> BaseCoreDataset[T_sample]:
    """
    Gets a dataset from a config path.

    Args:
        path: Path to the folder where the `.nv-meta` folder is contained.
        dataset_config: Filename of the dataset config file (`path / '.nv-meta' / config`)
        split_config: Filename of the split config file (`path / '.nv-meta' / split_config`)
        split_part: Name of the split to load.
        training: If true, apply training randomization and loop the dataset.
        subflavor: Override the __subflavor__ property of each sample.
        subflavors: Merge-Override the __subflavors__ property of each sample.
        worker_config: If set, use this worker config instead of the default one.
        sample_type: Type of the samples to load, only used to ensure typing.
        **kwargs: Additional arguments to be passed to the dataset constructor.

    Returns:
        The instantiated dataset
    """
    path = EPath(path).absolute()
    if not (path / MAIN_FOLDER_NAME).is_dir():
        raise ValueError(
            f"Path {path} does not contain a {MAIN_FOLDER_NAME} folder. Did you forget to prepare "
            f"the dataset? Please check the documentation for an introduction to dataset "
            f"preparation."
        )
    dataset = load_config(
        path / MAIN_FOLDER_NAME / dataset_config,
        default_kwargs=dict(
            path=path,
            split_config=split_config,
            split_part=split_part,
            training=training,
            subflavor=subflavor,
            worker_config=worker_config,
            **kwargs,
        ),
    )
    if dataset.subflavors is None:
        dataset.subflavors = subflavors
    elif subflavors is not None:
        dataset.subflavors.update(subflavors)
    if sample_type is not None:
        assert issubclass(
            dataset.__sample_type__, sample_type
        ), f"Sample of type {dataset.__sample_type__} is not a subclass of {sample_type}."
    return dataset



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import Sample, SavableDataset
from megatron.energon.rng import WorkerRng, WorkerRngMergedState, WorkerRngState
from megatron.energon.worker import WorkerConfig, global_worker_config
from megatron.energon.wrappers.base import (
    BaseSingleWrapperDataset,
    BaseSingleWrapperMergedState,
    BaseSingleWrapperState,
)

T_sample = TypeVar("T_sample")


@dataclass
class ShuffleBufferState(BaseSingleWrapperState):
    buffer: List[Tuple[Union[str, int], ...]]
    rng: WorkerRngState


@dataclass
class ShuffleBufferMergedState(BaseSingleWrapperMergedState):
    buffer: List[List[Tuple[Union[str, int], ...]]]
    rng: WorkerRngMergedState


class ShuffleBufferDataset(BaseSingleWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """Shuffle buffer for the dataset."""

    size: int
    worker_config: WorkerConfig
    _worker_rng: WorkerRng

    _active_buffer: List[List[T_sample]]
    _active_buffer_restore_keys: List[List[Tuple[Union[str, int], ...]]]
    _restore_pending: bool

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        size: int,
        worker_config: Optional[WorkerConfig] = None,
    ):
        """Create a shuffle buffer for the dataset."""
        super().__init__(dataset)
        self.size = size
        self.worker_config = worker_config or global_worker_config
        self._worker_rng = WorkerRng(self.worker_config)
        self._active_buffer = [[] for _ in range(max(self.worker_config.num_workers, 1))]
        self._active_buffer_restore_keys = [
            [] for _ in range(max(self.worker_config.num_workers, 1))
        ]
        self._restore_pending = False

    def __len__(self) -> int:
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample]:
        worker_idx = self.worker_config.rank_worker_id()
        # Cleanup other states
        for i in range(self.worker_config.num_workers):
            if i != worker_idx:
                self._active_buffer[i].clear()
                self._active_buffer_restore_keys[i].clear()
        active_buffer = self._active_buffer[worker_idx]
        active_buffer_restore_keys = self._active_buffer_restore_keys[worker_idx]
        if self._restore_pending:
            assert len(active_buffer) == 0
            self._restore_pending = False
            for restore_key in active_buffer_restore_keys:
                active_buffer.append(self.dataset.restore_sample(restore_key))
        assert len(active_buffer) == len(active_buffer_restore_keys)
        it = iter(self.dataset)
        while True:
            if len(active_buffer) >= self.size:
                pop_idx = self._worker_rng.randbelow(len(active_buffer))
                active_buffer_restore_keys.pop(pop_idx)
                yield active_buffer.pop(pop_idx)
            else:
                try:
                    sample = next(it)
                except StopIteration:
                    break
                else:
                    active_buffer.append(sample)
                    if isinstance(sample, Sample):
                        active_buffer_restore_keys.append(sample.__restore_key__)
                    elif isinstance(sample, dict) and "__restore_key__" in sample:
                        active_buffer_restore_keys.append(sample["__restore_key__"])
                    else:
                        active_buffer_restore_keys.append(())
        while len(active_buffer) > 0:
            pop_idx = self._worker_rng.randbelow(len(active_buffer))
            active_buffer_restore_keys.pop(pop_idx)
            yield active_buffer.pop(pop_idx)

    def save_state(self) -> ShuffleBufferState:
        from megatron.energon.flavors.crude import CrudeSample

        assert all(
            isinstance(el, (Sample, CrudeSample))
            for el in self._active_buffer[self.worker_config.rank_worker_id()]
        )
        assert (
            self.dataset.can_restore_sample()
        ), "Cannot restore sample from inner dataset, cannot save buffer state efficiently."
        return ShuffleBufferState.extend(
            super().save_state(),
            rng=self._worker_rng.save_state(),
            buffer=list(self._active_buffer_restore_keys[self.worker_config.rank_worker_id()]),
        )

    def merge_states(self, states: List[ShuffleBufferState]) -> ShuffleBufferMergedState:
        assert all(s is None or isinstance(s, ShuffleBufferState) for s in states)
        return ShuffleBufferMergedState.extend(
            super().merge_states(states),
            rng=self._worker_rng.merge_states([None if s is None else s.rng for s in states]),
            buffer=[[] if s is None else s.buffer for s in states],
        )

    def restore_state(self, state: Optional[ShuffleBufferMergedState]) -> None:
        super().restore_state(state)
        if state is None:
            self._active_buffer = [[] for _ in range(max(self.worker_config.num_workers, 1))]
            self._active_buffer_restore_keys = [
                [] for _ in range(max(self.worker_config.num_workers, 1))
            ]
            self._restore_pending = False
            self._worker_rng.restore_state(None)
        else:
            assert isinstance(state, ShuffleBufferMergedState)
            self._active_buffer = [[] for _ in range(max(self.worker_config.num_workers, 1))]
            self._active_buffer_restore_keys = state.buffer
            self._restore_pending = True
            self._worker_rng.restore_state(state.rng)

    def can_restore_sample(self) -> bool:
        return False

    def restore_sample(self, restore_key: Tuple[Union[str, int], ...]) -> T_sample:
        raise NotImplementedError("Cannot restore sample from shuffle buffer.")

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "size": self.size,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"ShuffleBufferDataset(size={self.size}, dataset={self.dataset})"



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import json
import multiprocessing
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Optional, TextIO

import torch.distributed
import torch.utils.data

from megatron.energon.epathlib import EPath

__all__ = (
    "WorkerConfig",
    "global_worker_config",
)


@dataclass(eq=False)
class WorkerConfig:
    """
    Provides information about the current worker and the global configuration. This gives each
    data parallel rank its proper config. Every `rank` (up to `world_size-1`) must be used.
    If set wrong, the datasets might yield the same data or data might be missing, as data
    is split over the data parallel ranks with this config!
    You may set the same rank, if you need multiple ranks to retrieve the same data.
    """

    #: The data parallel rank/id of the current process.
    rank: int
    #: The total number of data parallel processes.
    world_size: int
    #: The number of workers per rank. May be 0 to disable worker processes.
    num_workers: int

    #: If not using all ranks for data parallel, set this to the corresponding group.
    data_parallel_group: Optional[torch.distributed.ProcessGroup] = None

    #: The id offset of the current worker. e.g. the worker may live as `worker_info.id=0`, but
    # actually yield samples for id=1 (i.e. worker_id_offset=1). Required to support restoring the
    # worker state if last emitted sample was not for worker_id=0. Required by SavableDataLoader to
    # restore the worker state. Is only set to nonzero within a worker process.
    worker_id_offset: ClassVar[int] = 0

    #: The following seed_offset is used used at two points in the code.
    # 1. The seed_offset in the worker_config that is passed to the dataset initialization, is used
    #    to set the seed for the dataset shuffling and shuffled blending (All code that uses WorkerRng).
    # 2. The worker_config passed to the data loader initialization, is used to set the seed for the
    #    torch, numpy and random libraries. This does not affect the dataset shuffling, but only the
    #    user code (e.g. code in TaskEncoder).
    # Typically, you will not have to worry about this, and just use the same worker_config for both.
    # However, if you need to, you can pass two different worker_configs to the dataset and the
    # data loader.
    seed_offset: int = 0

    #: The path to the debug file for the current worker. Should contain "{worker_id}" and "{pid}"
    # to separate the workers.
    worker_debug_path: Optional[str] = None
    #: Log level for worker logging.
    worker_log_level: int = 0
    #: The opened file for the current worker. Should not be set from outside.
    _worker_debug_file: Optional[TextIO] = None
    #: worker_id of the opened worker debug file
    _worker_debug_file_worker_id: Optional[int] = None

    #: The current sample index within the current iterating worker
    _sample_index: ClassVar[Optional[int]] = None
    #: The current worker config within the current iterating worker
    active_worker_config: ClassVar[Optional["WorkerConfig"]] = None

    def worker_activate(self, sample_index: int):
        """Activates the worker config for the current worker and sets it as actively iterating.
        Must be called before next() call on the datasets."""
        WorkerConfig._sample_index = sample_index
        WorkerConfig.active_worker_config = self

    def worker_deactivate(self):
        """Deactivates the worker config for the current worker and deactivates it for iterating.
        Must be called after next() call on the datasets."""
        WorkerConfig._sample_index = None
        WorkerConfig.active_worker_config = None

    @property
    def active_worker_sample_index(self) -> int:
        """Returns the current sample index for the actively iterating worker."""
        # Internal sample index is for the local worker. If using multiple workers per rank, this
        # must be multiplied by the number of workers and offset by the local worker index.
        return WorkerConfig._sample_index * max(self.num_workers, 1) + self.rank_worker_id()

    def __eq__(self, other):
        """Do not compare everything to check for equal config"""
        if not isinstance(other, WorkerConfig):
            return NotImplemented
        return all(
            [
                self.rank == other.rank,
                self.world_size == other.world_size,
                self.num_workers == other.num_workers,
            ]
        )

    @staticmethod
    def default_worker_config(
        num_workers: int = 4, data_parallel_group: Optional[torch.distributed.ProcessGroup] = None
    ) -> "WorkerConfig":
        """Returns the default worker config using torch distributed."""
        if torch.distributed.is_available() and torch.distributed.is_initialized():
            rank = torch.distributed.get_rank(data_parallel_group)
            world_size = torch.distributed.get_world_size(data_parallel_group)
        else:
            rank = 0
            world_size = 1
        return WorkerConfig(
            rank=rank,
            world_size=world_size,
            num_workers=num_workers,
            data_parallel_group=data_parallel_group,
        )

    def rank_worker_id(self) -> int:
        """Returns the self worker id within the current rank."""
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            return self.worker_id_offset
        return (
            worker_info.id + worker_info.num_workers - self.worker_id_offset
        ) % worker_info.num_workers

    def assert_worker(self):
        """Checks if the current process is a worker (if configured so), and that the workers are
        properly configured."""
        if self.num_workers <= 1:
            assert self.rank_worker_id() == 0
        else:
            worker_info = torch.utils.data.get_worker_info()
            assert worker_info is not None, "Cannot iterate out of worker context"
            assert worker_info.num_workers == self.num_workers, (
                f"Actual number of workers for this rank ({worker_info.num_workers}) does not "
                f"match the configured number of workers ({self.num_workers})"
            )

    def global_worker_id(self, override_local_worker_id: Optional[int] = None) -> int:
        """Returns the global worker index by multiplying the rank with the number of workers.
        Alternatively, you can override the local worker id.

        Args:
            override_local_worker_id (int, optional): The local worker id to override. None means
                the current worker, which is the default.
        """

        if override_local_worker_id is not None:
            return self.rank * self.num_workers + override_local_worker_id
        else:
            self.assert_worker()
            return self.rank * self.num_workers + self.rank_worker_id()

    def worker_seed(self, override_local_worker_id: Optional[int] = None) -> int:
        """Returns the seed for the current worker (or a specified worker).
        Base on the current worker id and the seed offset, compute a seed.
        Alternatively, you can override the local worker id with a fixed one to
        pregenerate seeds for multiple workers.

        Args:
            override_local_worker_id (int, optional): The local worker id to override. None means
                the current worker, which is the default.
        """

        if self.num_workers == 0:
            # If we are not using workers, different ranks should still get a different seed
            global_worker_id = self.rank
        else:
            global_worker_id = self.global_worker_id(override_local_worker_id)

        seed_offset = self.seed_offset

        seed_hash = hashlib.sha1(f"{global_worker_id},{seed_offset}".encode("utf-8")).digest()

        return int.from_bytes(seed_hash, byteorder="big", signed=False) & 0xFFFFFFFF

    def config(self) -> Dict[str, Any]:
        return {
            "rank": self.rank,
            "world_size": self.world_size,
            "num_workers": self.num_workers,
            "data_parallel_group": (
                self.data_parallel_group.size() if self.data_parallel_group else None
            ),
        }

    def should_log(self, level: int) -> bool:
        return level <= self.worker_log_level

    def worker_log(self, data: dict) -> None:
        """Logs the given data to the worker debug file."""
        if self.worker_debug_path is None:
            print(json.dumps(data) + "\n", end="", flush=True)
        else:
            in_worker = torch.utils.data.get_worker_info() is not None
            # Additional "worker" with rank_worker_id=0 is the main process. All workers have +1
            # as their worker_id.
            worker_id = (
                self.rank * (self.num_workers + 1) + self.rank_worker_id() + (1 if in_worker else 0)
            )
            if self._worker_debug_file is None or self._worker_debug_file_worker_id != worker_id:
                if self._worker_debug_file is not None:
                    self._worker_debug_file.close()
                path = EPath(
                    self.worker_debug_path.format(
                        worker_id=worker_id, pid=multiprocessing.current_process().ident
                    )
                )
                path.parent.mkdir(exist_ok=True, parents=True)
                self._worker_debug_file = path.open("w")
                self._worker_debug_file_worker_id = worker_id
            self._worker_debug_file.write(json.dumps(data) + "\n")
            self._worker_debug_file.flush()


#: The global worker config instance used if not overridden.
global_worker_config = WorkerConfig.default_worker_config(num_workers=0)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import struct
import tarfile
from types import TracebackType
from typing import BinaryIO, Optional, Type, Union

from megatron.energon.epathlib import EPath
from megatron.energon.retry_stream import RetryReadStream

ITAR_SUFFIX = ".tar.idx"


class TarIndexReader:
    def __init__(self, tar_path: Union[EPath, str]):
        tar_path = EPath(tar_path)
        self.itar = (tar_path.with_suffix(ITAR_SUFFIX)).open("rb")
        self._length = len(self)

    def __getitem__(self, index: int) -> int:
        if index >= self._length or index < 0:
            raise IndexError(f"Index {index} out of range")

        self.itar.seek(8 * index)
        return struct.unpack("Q", self.itar.read(8))[0]

    def __len__(self) -> int:
        self.itar.seek(0, 2)
        return self.itar.tell() // 8

    def close(self):
        self.itar.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TarIndexWriter:
    def __init__(self, tar_path: EPath):
        self.final_name = tar_path.with_suffix(ITAR_SUFFIX)
        self.tmp_name = tar_path.with_suffix(ITAR_SUFFIX + ".tmp")
        self.itar = self.tmp_name.open("wb")

    def append(self, offset: int):
        self.itar.write(struct.pack("Q", offset))

    def close(self, finalize: bool = True):
        self.itar.close()
        if finalize:
            self.tmp_name.move(self.final_name)
        else:
            self.tmp_name.unlink()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(finalize=exc_val is None)


class SubFileReader(BinaryIO):
    """A file-like object that reads a subfile (i.e. offset, size defined portion) of a larger
    file."""

    def __init__(self, stream: BinaryIO, offset: int, size: int):
        self.offset = offset
        self._pos = 0
        self.size = size
        self.stream = stream
        self.stream.seek(self.offset)

    def read(self, n: int = -1) -> bytes:
        if n == -1:
            n = self.size - self._pos
        else:
            n = min(n, self.size - self._pos)
        if n == 0:
            return b""
        read = self.stream.read(n)
        self._pos += len(read)
        return read

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos += offset
        elif whence == 2:
            self._pos = self.size + offset
        else:
            raise ValueError("Invalid whence value")
        self._pos = max(0, min(self._pos, self.size))
        self.stream.seek(self.offset + self._pos)
        return self._pos

    def tell(self) -> int:
        return self._pos

    def __enter__(self) -> BinaryIO:
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        self.close()

    def close(self) -> None:
        self.stream.close()

    def isatty(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False


def get_itar_byte_offset(
    path: Union[str, EPath],
    sample_offset: int = 0,
) -> int:
    """Gets the byte offset from sample offsets."""
    if sample_offset == 0:
        return 0
    with TarIndexReader(path) as itar:
        return itar[sample_offset]


@contextlib.contextmanager
def open_itar(path: Union[str, EPath], byte_offset: int = 0, byte_size: Optional[int] = None):
    """
    Open an indexed tarfile with offset and size.
    Args:
        path: Path to the tarfile to open
        byte_offset: Byte offset within the file
        byte_size: Size of the file to read

    Returns:
        The opened tarfile
    """
    path = EPath(path)

    # TODO: if tar file startswith(b"\x1f\x8b\x08") -> Seekable gzip file
    with path.open("rb") as f:
        if f.read(3) == b"\x1f\x8b\x08":
            # Open as seekable tgz
            raise ValueError("Seekable tgz not supported yet")

    if byte_offset != 0 or byte_size is not None:
        if byte_size is None:
            byte_size = path.size() - byte_offset
        with RetryReadStream(path) as stream:
            with SubFileReader(
                stream,
                offset=byte_offset,
                size=byte_size,
            ) as fileobj:
                with tarfile.open(fileobj=fileobj, mode="r|") as f:
                    yield f
    else:
        with RetryReadStream(path) as fileobj:
            with tarfile.open(fileobj=fileobj, mode="r|") as f:
                yield f



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import traceback


def log_exception(_e: Exception, sample, /):
    traceback.print_exc()
    print("-" * 10)

    sample_str = str(sample)
    if len(sample_str) > 400:
        sample_str = sample_str[:200] + "..." + sample_str[-200:]

    print(sample_str)

    print("-" * 10)



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import List, Optional, Union

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class OCRSample(Sample):
    """Sample type for optical character recognition."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor
    #: The text contained in the image
    text: str
    #: The bounding boxes of the blocks in the image float(N, 4|5<x, y, w, h>)
    block_boxes: Optional[torch.Tensor] = None
    #: The classes of the blocks in the image int(N, 1<block_class>)
    block_classes: Optional[Union[torch.Tensor, List[str]]] = None
    #: The text contained in each block (N,)
    block_text: Optional[List[str]] = None
    #: The bounding boxes of the lines in the image float(N, 4|5<x, y, w, h[, confidence]>)
    lines_boxes: Optional[torch.Tensor] = None
    #: The text contained in each line (N,)
    lines_text: Optional[List[str]] = None
    #: The bounding boxes of the words in the image float(N, 4|5<x, y, w, h[, confidence]>)
    words_boxes: Optional[torch.Tensor] = None
    #: The text contained in each word (N,)
    words_text: Optional[List[str]] = None
    #: The bounding boxes of the chars in the image float(N, 4|5<x, y, w, h[, confidence]>)
    chars_boxes: Optional[torch.Tensor] = None
    #: The character contained in each char (N,)
    chars_text: Optional[List[str]] = None


class OCRWebdataset(DefaultDecoderWebdataset[OCRSample]):
    __sample_type__ = OCRSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class CaptioningSample(Sample):
    """Sample type for image captioning."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor

    #: The caption string
    caption: str


class CaptioningWebdataset(DefaultDecoderWebdataset[CaptioningSample]):
    __sample_type__ = CaptioningSample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.metadataset.dataset_loader import DatasetLoader
from megatron.energon.metadataset.loader import load_dataset
from megatron.energon.metadataset.loader_interface import DatasetLoaderInterface
from megatron.energon.metadataset.metadataset import Metadataset



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import List, Optional

import torch

from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.base_webdataset import DefaultDecoderWebdataset


@dataclass
class VQASample(Sample):
    """Sample type for visual question answering."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor
    #: The context/question for the image
    context: str

    #: The possible answers. Not set for testing.
    answers: Optional[List[str]] = None
    #: The weights of the possible answers. Optionally available.
    answer_weights: Optional[torch.Tensor] = None


class VQAWebdataset(DefaultDecoderWebdataset[VQASample]):
    __sample_type__ = VQASample



# Copyright (c) 2024, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Union

import webdataset

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.flavors.base_webdataset import (
    BaseWebdataset,
    DefaultGenericWebdataset,
    ImageDecoder,
)
from megatron.energon.flavors.webdataset.structs import FilteredSample
from megatron.energon.wrappers import MapDataset
from megatron.energon.wrappers.iter_map_dataset import IterMapDataset


class CrudeSample(dict):
    """Generic sample type to be processed later."""


class CrudeWebdataset(DefaultGenericWebdataset[CrudeSample]):
    """The CrudeWebdataset is used to load crude / raw samples and
    decode them in the user code using so-called cookers.

    See the documentation under "Crude Data" for more information.
    """

    __sample_type__ = CrudeSample

    def __init__(
        self,
        path: EPath,
        *,
        subflavor: Optional[str] = None,
        subflavors: Optional[Dict[str, Any]] = None,
        part_filter: Optional[Union[str, List[str], Callable[[str], bool]]] = None,
        image_decode: Optional[ImageDecoder] = "torchrgb",
        ignore_decoder_errors: bool = False,
        **kwargs,
    ):
        # The sample is already a dict, but we extend it with the subflavor
        self.subflavor = subflavor
        self.subflavors = subflavors or {}
        self.image_decode = image_decode
        self.ignore_decoder_errors = ignore_decoder_errors

        self._sample_loader = lambda sample: {
            **sample,
            "__subflavor__": self.subflavor,
            "__subflavors__": self.subflavors,
        }

        # A simple part filter can be specified as a list of accepted extensions
        if isinstance(part_filter, list):
            parts = set(part_filter)
            part_filter = lambda part: part in parts
        else:
            assert part_filter is None

        # We skip the parent class __init__ and call the BaseWebdataset.__init__ directly
        BaseWebdataset.__init__(self, path, **kwargs, part_filter=part_filter)

    def _decode_error_handler(self, exc: Exception) -> bool:
        if self.ignore_decoder_errors:
            return True
        raise exc

    def _process_samples(
        self, dataset: SavableDataset[FilteredSample]
    ) -> SavableDataset[CrudeSample]:
        if self.image_decode:
            # With image decoding:
            decoder = webdataset.decode(self.image_decode, handler=self._decode_error_handler)
            dataset = IterMapDataset(
                dataset, decoder, error_handler=self.error_handler, stateless_iter_fn=True
            )

        return MapDataset(
            dataset, self._load_sample, error_handler=self.error_handler, stateless_map_fn=True
        )

    def _load_sample(self, sample: FilteredSample) -> CrudeSample:
        return self.__sample_type__(**self._sample_loader(sample))

    def config(self) -> Dict[str, Any]:
        return {
            **super().config(),
            "image_decode": self.image_decode,
            "ignore_decoder_errors": self.ignore_decoder_errors,
        }
